# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "optparse"
$LOAD_PATH << File.join(File.dirname($PROGRAM_NAME), "..", "lib")

require "noms"

def proccessing_msg(path)
  puts "Processing #{path}"
end

def open_if_exist(path, file_klass=File, &blk)
  if File.exist?(path)
    file_klass.open(path, &blk)
  end
end

def main
  if NOMS.options[:profile]
    require "ruby-prof"
    RubyProf.start
  end
  
  puts "Using configuration for #{NOMS.config.name}"
  start_time = Time.now
  data = Hash.new { |h, name| h[name] = NOMS::Light::Player.new(name) }
  
  var_dir = NOMS::Setup.setup_var_dir
  ladderlog_path = File.join(var_dir, "ladderlog.txt")
  won_matches_path = File.join(var_dir, "won_matches.txt")
  won_rounds_path = File.join(var_dir, "won_rounds.txt")
  ladder_path = File.join(var_dir, "ladder.txt")
  
  should_process_stats = true
  open_if_exist(ladderlog_path, NOMS::Util::ReverseFile) do |f|
    should_process_stats = NOMS::Ladderlog.should_process_ladderlog?(f)
    if should_process_stats
      proccessing_msg(ladderlog_path)
      matches, ladderlog, match_number = NOMS::Ladderlog.process_ladderlog(f, data)
    
      if NOMS.options[:partial_update]
        proccessing_msg(ladderlog_path + " for additional information about players")
        NOMS::Ladderlog.process_player_information(f, data)
      end
    
      if NOMS.config.armaskill.generate
        proccessing_msg("ArmaSkill ratings")
        NOMS::ArmaSkill.process_skills(matches, data)
      end
    
      NOMS::Setup.save_meta_ladderlog(ladderlog, match_number, f)
    end
  end
  
  if !should_process_stats
    puts "Not processing stats. They have not changed since last update"
    return
  end
  
  if !NOMS.options[:development_mode]
    meta = NOMS::DB::Meta.first
    
    open_if_exist(won_matches_path) do |f|
      proccessing_msg(won_matches_path)
      NOMS::StatFile.process_stat_file(meta, :old_won_matches_txt, f.read, data, :won_matches) { |won| Integer(won) }
    end
    
    open_if_exist(won_rounds_path) do |f|
      proccessing_msg(won_rounds_path)
      NOMS::StatFile.process_stat_file(meta, :old_won_rounds_txt, f.read, data, :won_rounds) { |won| Integer(won) }
    end
  
    open_if_exist(ladder_path) do |f|
      proccessing_msg(ladder_path)
      NOMS::StatFile.process_stat_file(meta, :old_ladder_txt, f.read, data, :ladder) { |won| Float(won) }
    end
    
    meta.save_changes
  end
  
  puts "Exporting #{data.length} player entries from #{NOMS.config.name}"
  NOMS::Setup.save(data)
  
  puts "Moving temporary database to #{NOMS::Setup.database_file(false)}" if !NOMS.options[:partial_update]
  NOMS::Setup.finish(NOMS.options[:partial_update])
  
  puts "Processing took %.2f seconds" % (Time.now - start_time)
  
  if NOMS.options[:profile]
    profile_result = RubyProf.stop
    printer = RubyProf::GraphHtmlPrinter.new(profile_result)
    File.open("noms_profile.html", "w") do |f|
      printer.print(f, :min_percent=>0)
    end
  end
end

def parse_options
  options = {}
  parser = OptionParser.new do |opts|
    opts.banner = "Usage: ruby #{$PROGRAM_NAME} [options] <noms.config>"
    
    opts.on("--matches-limit N", "Parse upto N matches from ladderlog.txt.") do |n|
      options[:matches_limit] = Integer(n)
    end
    
    opts.on("--db-directory DIR", "The directory to store databases in. Default=#{NOMS.options[:db_directory]}") do |dir|
      NOMS.options[:db_directory] = dir
    end
    
    opts.on("--development-mode", "") do
      NOMS.options[:development_mode] = true
    end
    
    opts.on("--profile", "Generates a HTML profile report.") do
      NOMS.options[:profile] = true
    end
    
    opts.on("--partial-update", "Partially update the stats, instead of doing a full parse.") do
      NOMS.options[:partial_update] = true
    end
  end
  
  begin
    parser.parse!
  rescue OptionParser::InvalidOption => e
    puts parser.banner
    puts "#{e.class}: #{e}"
    exit(1)
  end
  
  return [parser, options]
end

if __FILE__ == $PROGRAM_NAME
  parser, options = parse_options()
  
  if ARGV.length != 1
    puts parser.banner
    puts "Error: No configuration file specified."
    exit(1)
  end
  
  NOMS.options.merge!(options)
  
  config_path = ARGV[0]
  
  DB = NOMS::Setup.add_config(config_path, !NOMS.options[:partial_update])
  NOMS::Setup.generate_models(!NOMS.options[:partial_update])
  NOMS.current_db.synchronous = :off
  
  main()
end
