# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

module NOMS
  module Achievement
    
    class BaseAchievement
      include Util::RenameHelper
      
      def initialize(match, light_class, add_method)
        @match = match
        @light_class = light_class
        @add_method = add_method
      end
      attr_reader :match
      attr_reader :add_method
      
      def to_light(match_number, number_matches_won)
        achievement = @light_class.new
        achievement.global_match_number = match_number
        achievement.player_match_win_number = number_matches_won
        achievement
      end
      
      def all_rounds_played?(player)
        self.match.players[player].rounds_played >= self.match.number_rounds
      end
    end

    class PerfectMatch < BaseAchievement
      def initialize(match)
        @died = Hash.new { |h, k| h[k] = false }
        super(match, NOMS::Light::PerfectMatchAchievement, :add_perfect_match_achievement)
      end
      
      def death_suicide(action)
        @died[action.player] = true
      end
      
      def death_frag(action)
        @died[action.prey] = true
      end
      
      def death_teamkill(action)
        @died[action.prey] = true
      end
            
      def player_renamed(action)
        rename_helper(@died, action.old_name, action.new_name)
      end
      
      def accomplished?(player)
        all_rounds_played?(player) && !@died[player]
      end
    end

    class MinimumMatch < BaseAchievement
      def initialize(match)
        super(match, NOMS::Light::MinimumMatchAchievement, :add_minimum_match_achievement)
      end
      
      def accomplished?(player)
        Ladderlog::MINIMUM_ROUNDS_NEEDED_FOR_MATCH > 0 && self.match.number_rounds == Ladderlog::MINIMUM_ROUNDS_NEEDED_FOR_MATCH
      end
    end
  
    # Win a match, and don't win a round
    class MatchWinnerRoundLoser < BaseAchievement
      def initialize(match)
        @round_winner = Hash.new { |h, k| h[k] = false }
        super(match, NOMS::Light::MatchWinnerRoundLoserAchievement, :add_match_winner_round_loser_achievement)
      end

      def round_winner(action)
        self.match.win_players(action).each do |player|
          @round_winner[player] = true
        end
      end
      
      def player_renamed(action)
        rename_helper(@round_winner, action.old_name, action.new_name)
      end

      def accomplished?(player)
        return all_rounds_played?(player) && !@round_winner[player]
      end      
    end
    
    # Win a match, and win every round
    class MatchWinnerRoundWinner < BaseAchievement
      def initialize(match)
        @round_wins = Hash.new { |h, k| h[k] = 0 }
        super(match, NOMS::Light::MatchWinnerRoundWinnerAchievement, :add_match_winner_round_winner_achievement)
      end
      
      def round_winner(action)
        self.match.win_players(action).each do |player|
          @round_wins[player] += 1
        end
      end
      
      def player_renamed(action)
        rename_helper(@round_wins, action.old_name, action.new_name)
      end
      
      def accomplished?(player)
        all_rounds_played?(player) && self.match.players[player].rounds_played == @round_wins[player]
      end
    end
    
    class Domination < BaseAchievement
      def initialize(match)
        super(match, NOMS::Light::DominationAchievement, :add_domination_achievement)
      end
      
      def accomplished?(winner)
        winner_team = self.match.players[winner].team
        # Need to check players instead of teams, because the team might have dissolved.
        all_rounds_played?(winner) && !self.match.players.values.any? { |player_data| player_data.team != winner_team && player_data.score > 0  }
      end
    end
    
  end
end
