# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2010  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "rubygems"
require "sequel"
require "rexml/document"
require "noms/util"
require "fileutils"

module NOMS
  module Choropleth
    DB_PATH = "db/countries.sqlite3"
    
    class << self
      attr_accessor :countries_db
    end
    
    def self.available?
      @__available = File.exist?(Choropleth::DB_PATH) if @__available.nil?
      @__available
    end
        
    def self.connect_db
      FileUtils.mkdir_p(File.dirname(DB_PATH))
      Choropleth.countries_db = Sequel.connect("sqlite://#{DB_PATH}")
    end
    
    def self.setup_db(should_create_tables)
      if should_create_tables
        Choropleth.countries_db.create_table(:entries) do
          primary_key(:id)
          column(:ip_from, :integer)
          column(:ip_to, :integer)
          column(:country_code, :string)
          column(:country, :string)
        end
      end
      require "noms/models/choropleth"
      Choropleth::Entry.db = Choropleth.countries_db
    end
    
    def self.import(path)
      FasterCSV.foreach(path) do |row|
        entry = Entry.new
        entry.ip_from = row[0].to_i
        entry.ip_to = row[1].to_i
        entry.country_code = row[4]
        entry.country = row[6]
        entry.save
      end
    end
    
    class Map
      COLORS = [
        "rgb(254, 217, 118)",
        "rgb(254, 178, 76)",
        "rgb(253, 141, 60)",
        "rgb(252, 78, 42)",
        "rgb(227, 26, 28)",
        "rgb(177, 0, 38)" 
      ]
      
      WORLD_MAP_PATH = File.join(File.dirname(__FILE__), "world-map.svg")
      
      def initialize
        @countries = determine_country_colors()
      end
      
      attr_reader :countries
      
      def to_json
        @countries.each do |id, c|
          c["name"] = CountryCode.country_name_from_alpha2(id)
          flag = "flags/#{id}.png"
          if File.exist?(NOMS::Util.public_path(flag))
            c["flag"] = "/#{flag}"
          else
            c["flag"] = ""
          end
        end
        JSON.dump(@countries)
      end
      
      protected
      
      def determine_country_colors
        counts = NOMS::DB::Player.countries_simple_counts
        sorted_counts = counts.sort_by { |a2, count| count }
        just_counts = sorted_counts.map { |(a2, count)| count }
        bin_boundaries = 0.enum_for(:step, 1.0, 1.0 / (COLORS.length - 1)).inject([]) { |memo, n| memo << quantile(just_counts, n) }.reverse
        bins = Hash.new { |h, k| h[k] = [] }
        sorted_counts.each do |(a2, n)|
          boundary_index = bin_boundaries.enum_for(:each_with_index).find { |boundary, index| n >= boundary }[1]
          bins[COLORS[COLORS.length - boundary_index - 1]] << a2
        end
        bins.inject({}) { |memo, (color, countries)|
          countries.each { |c| memo[c] = {"color" => color, "count" => counts[c]} }
          memo
        }
      end
      
      def modf(n)
        i = n.floor
        [n - i, i.to_f]
      end

      def quantile(x, prob)
        a, b, c, d = 1.0, -1.0, 0.0, 1.0
        n = x.length
        g, j = modf(a + (n + b) * prob - 1)
        if j < 0
          return x[0]
        elsif j > n
          return x[n]
        end

        j = j.floor
        return x[j] if g == 0
        return x[j] + (x[j + 1] - x[j]) * (c + d * g)
      end      
    end
    
  end
end
