# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "yaml"

module NOMS
  class << self
    attr_accessor :config
  end
  
  class Config
    class Error < StandardError; end
    
    class Validator
      def initialize(optional, description, default, &blk)
        @optional = optional
        @description = description
        @default = default
        @validator = blk
      end
      
      attr_reader :default
      attr_reader :description
      
      def optional?
        @optional
      end
      
      def valid?(value)
        @validator.call(value)
      end
    end
    
    def initialize
      @valid_options = []
      @validators = {}
      @options = {}
    end
    
    def new_nested_option(name, description)
      add_valid_option(name, description)
      @options[name] = Config.new
      yield @options[name] if block_given?
      @options[name]
    end
        
    def from_file(path, should_ensure_all_valid=true)
      conf = File.open(path) { |f| YAML.load(f) }
      from_file(File.join(File.dirname(path), conf["base_config"]), false) if conf["base_config"]
      conf.delete("base_config")
      from_hash(conf, should_ensure_all_valid)
    end
    
    def from_hash(conf, should_ensure_all_valid=true)
      import_options(conf)
      ensure_all_valid() if should_ensure_all_valid
    end
    
    def add_valid_option(name, description, default=nil, &blk)
      @valid_options << name
      blk = lambda { |_| true } if !blk
      @validators[name] = Validator.new(!default.nil?, description, default, &blk)
    end
    
    def [](key)
      if is_option?(key)
        return option_value(key)
      end
      raise Error.new("Option “#{key}” is not a valid option.")
    end
  
    def define_config
      self.add_valid_option("name", "The name of your server.") { |value| value && value.any? }
      self.add_valid_option("short_name", "The short name for your server.") { |value| value && value.any? }
      self.add_valid_option("var_directory", "The var directory of your server which contains all the statistic files.", "")
      self.add_valid_option("var_directory_uri", "A URI to fetch a .tar.gz file that contains the files in var directory.", "")
      self.add_valid_option("group", "The group this server belongs to", "")
      self.add_valid_option("servers_grouped", "The short names of the servers grouped for this server", [])
      self.add_valid_option("show_expanded_datasets", "Show pages like the Match History page or the Achievements page.", true)
      
      self.new_nested_option("armaskill", "Options for the armaskill program.") do |c|
        validator = lambda { |value| !!Float(value) }
        c.add_valid_option("generate", "Generate ratings for each player.", false)
        c.add_valid_option("stddev-multiplier", "", 0, &validator)
        c.add_valid_option("mean","",  0, &validator)
        c.add_valid_option("beta", "", 0, &validator)
        c.add_valid_option("dynamics-factor", "", 0, &validator)
        c.add_valid_option("draw-probability", "", 0, &validator)
      end
      
      self.new_nested_option("sanity_checks", "Sanity checks used when determining the validity of events.") do |c|
        validator = lambda { |value| !!Integer(value) }
        c.add_valid_option("limit_score", "The score limit for the match. Example: 100 in Fortress.", 0, &validator)
        c.add_valid_option("max_round_score", "The maximum number of points a person can acquire in one round.", 0, &validator)
        c.add_valid_option("min_round_score", "On sty+ct, ROUND_SCORE is broken and will sometimes be negative after going into spectator mode.", "default")
        c.add_valid_option("minimum_rounds_needed_for_match", "The minimum number of rounds it takes a player to win a match.", 0, &validator)
        c.add_valid_option("ladderlog_complete", "Special case for incomplete ladderlogs", true)
        c.add_valid_option("ladderlog_has_ip_addresses", "Special case for ladderlogs which have IP addresses removed.", true)
        c.add_valid_option("sty_ct", "This server uses sty+ct", false)
        c.add_valid_option("has_simple_stats", "Use ladderlog won matches/rounds for display if false", true)
        c.add_valid_option("filter_no_rounds_played", "Display stats from players with no rounds played", false)
        c.add_valid_option("ladderlog_utf8", "The ladderlog is in the UTF8 encoding", false)
      end
      
      # self.new_nested_option("achievements", "TODO add descriptions") do |c|
      #   c.add_valid_option("domination_match", "", true)
      #   c.add_valid_option("min_match", "", true)
      #   c.add_valid_option("perfect_match", "", true)
      #   c.add_valid_option("round_loser_match", "", true)
      #   c.add_valid_option("round_winner_match", "", true)
      # end
      
      define_stats_config("stats_table", true)
      define_stats_config("ratings_stats_table", false)
      define_stats_config("compare_stats_table", false)
    end
    
    def with_each(level=0, &blk)
      @valid_options.each do |name|
        if is_nested_option?(name)
          yield(name, @validators[name].description, level)
          @options[name].with_each(level + 1, &blk)
        else
          yield(name, @validators[name].description, level)
        end
      end
      
    end
    
    protected
    
    def define_stats_config(name, should_enable_special)
      self.new_nested_option(name, "Options for #{name}") do |c|
        stats_table = NOMS::WebApp::StatsTable.new(false)
        stats_table.groups.each do |group|
          c.new_nested_option(group.name, "The stats table group for #{group.name}") do |group_config|
            group_config.add_valid_option("enabled", "Should this entire group be enabled?", "default") { |_| true }
            group_config.add_valid_option("available", "Should this entire group be available to the user?", true) if should_enable_special
            group.options.each do |opt|
              group_config.add_valid_option(opt.name, opt.description, "default")
            end
          end
        end
      end
    end
    
    def import_options(options)
      options.each do |key, value|
        if value.is_a?(Hash)
          @options[key].import_options(value)
        else
          raise Error.new("Option “#{key}” is not a valid option.") if !is_option?(key)
          @options[key] = value
        end
      end
      
    end
    
    def ensure_all_valid
      @valid_options.each do |name|
        if is_nested_option?(name)
          @options[name].ensure_all_valid
        else
          ensure_option_valid(name)
        end
      end
    end
    
    def []=(key, value)
      # puts "Setting #{key.inspect} as #{value.inspect}"
      @options[key] = value
    end
        
    def is_option?(name)
      @valid_options.include?(name)
    end
    
    def is_nested_option?(name)
      return @options[name].is_a?(Config)
    end
    
    def is_option_set?(name, strict_check=false)
      
      # It's a nested group, such as config.foo.bar
      return true if is_nested_option?(name)
      
      validator = validator_for(name)
      raise Error.new("Option “#{name}” does not have a validator.") unless validator
      
      if validator.optional? && !@options.include?(name)
        return !strict_check
      end
          
      return true if validator.optional?
      
      return true if @options.include?(name)
      
      return false
    end
    
    def is_option_valid?(name)
      if is_option_set?(name)
        return true if is_nested_option?(name)
        validator = validator_for(name)
        return true if !is_option_set?(name, true)
        return validator.valid?(@options[name])
      end
      false
    end
    
    def validator_for(name)
      @validators[name]
    end
    
    def option_value(name)
      validator = validator_for(name)
      if is_nested_option?(name) || is_option_set?(name, true)
        return @options[name]
      elsif validator.optional?
        return validator.default
      end
      raise Error.new("This section should never be reached.")
    end
    
    def ensure_option_set(name)
      raise Error.new("Option “#{name}” is not optional, it must be set.") unless is_option_set?(name)
    end
    
    def ensure_option_valid(name)
      raise Error.new("Option “#{name}” is not valid.") unless is_option_valid?(name)
    end
        
    alias_method(:old_method_missing, :method_missing)
    def method_missing(name_sym, *args, &blk)
      name = name_sym.to_s
      if is_option?(name)
        return option_value(name)
      end
      old_method_missing(name_sym, *args, &blk)
    end

  end
end
