# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2010  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "noms/ladderlog/action"
require "enumerator"

module NOMS
  module Ladderlog
    
    # Alias these config items because we access them often.
    def self.alias_config_items
      configs = {
        :MINIMUM_ROUNDS_NEEDED_FOR_MATCH => NOMS.config.sanity_checks.minimum_rounds_needed_for_match,
        :MAX_ROUND_SCORE => NOMS.config.sanity_checks.max_round_score,
        :MIN_ROUND_SCORE => NOMS.config.sanity_checks.min_round_score,
        :LIMIT_SCORE => NOMS.config.sanity_checks.limit_score,
        :HAS_IP_ADDRESSES => NOMS.config.sanity_checks.ladderlog_has_ip_addresses,
        :STY_CT => NOMS.config.sanity_checks.sty_ct
      }
      configs.each do |constant, value|
        const_set(constant, value)
      end
    end
        
    class Match
      module Events
        def match_end
          process_round_scores()
          notify_achievements(:match_end)
        end

        def new_match
          process_round_scores()
          notify_achievements(:new_match)
        end

        def new_round
          @number_rounds += 1
          process_round_scores()
          @current_round = Round.new
          notify_achievements(:new_round)
        end

        def round_score(action)
          @current_round.round_scores << action if action.player && action.player.any?
          notify_achievements(:round_score, action)
        end

        def round_score_team(action)
          @current_round.round_scores_team << action
          notify_achievements(:round_score_team, action)
        end

        def death_suicide(action)
          @current_round.death_suicides << action
          notify_achievements(:death_suicide, action)
        end

        def death_frag(action)
          @current_round.death_frags << action
          notify_achievements(:death_frag, action)
        end

        def death_teamkill(action)
          @current_round.death_teamkills << action
          notify_achievements(:death_teamkill, action)
        end

        def match_winner(action)
          @winner = action
          notify_achievements(:match_winner, action)
        end

        def round_winner(action)
          @current_round.round_winner = action
          notify_achievements(:round_winner, action)
        end

        def player_renamed(action)
          return if action.old_name == action.new_name || action.new_name.nil?

          player_data = rename_helper(@players, action.old_name, action.new_name)
          is_authenticating = action.new_name.index("@")

          if player_data.team.any?
            @teams[player_data.team].players.delete(action.old_name)
            @teams[player_data.team].players << action.new_name
          end

          if player_data.team.any? && player_data.team == action.old_name && !is_authenticating
            team_data = rename_helper(@teams, action.old_name, action.new_name)
            team_data.name = action.new_name
          end
          
          @data[action.new_name].update_screen_name(action.screen_name, action.ip_address)
          notify_achievements(:player_renamed, action)
        end

        def player_entered(action)
          @data[action.player].update_screen_name(action.screen_name, action.ip_address)
          notify_achievements(:player_entered, action)
        end

        def player_left(action)
          @player_lefts << action
          notify_achievements(:player_left, action)
        end

        def basezone_conquered(action)
          @current_round.basezones_conquered << action
          notify_achievements(:basezone_conquered, action)
        end

        def basezone_conquerer(action)
          @current_round.basezones_conquered[-1].conquerers << action
          notify_achievements(:basezone_conquerer, action)
        end

        def sacrifice(action)
          @current_round.sacrifices << action
          notify_achievements(:sacrifice, action)
        end

        def cycle_created(action)
          @current_round.cycles_created << action
          notify_achievements(:cycle_created, action)
        end

        def targetzone_player_enter(action)
          @current_round.targetzone_player_enters << action
          notify_achievements(:targetzone_player_enter, action)
        end
        
        protected
        
        def notify_achievements(method, *args)
          @achievements.each do |achievement|
            achievement.method(method).call(*args) if achievement.respond_to?(method)
          end
        end
      end

      class PlayerData
        def initialize(name)
          @name = name
          @team = ""
          @rounds_played = 0
          @score = 0
        end

        attr_accessor(:name, :team, :rounds_played, :score)
      end

      class TeamData
        def initialize(name)
          @name = name
          @players = []
          @rounds_played = 0
          @score = 0
        end

        attr_accessor(:name, :players, :rounds_played, :score)
      end

      class Round
        def initialize
          @round_scores = []
          @round_scores_team = []
          @death_suicides = []
          @death_frags = []
          @death_teamkills = []
          @round_winner = nil
          @basezones_conquered = []
          @sacrifices = []
          @cycles_created = []
          @targetzone_player_enters = []
        end

        attr_accessor(
          :round_scores,
          :round_scores_team,
          :death_suicides,
          :death_frags,
          :death_teamkills,
          :round_winner,
          :basezones_conquered,
          :sacrifices,
          :cycles_created,
          :targetzone_player_enters
        )

        def valid_round?
          @round_scores.any?
        end

        def round_score_team_feature?
          @round_scores_team.any?
        end
      end

      def initialize
        @data = Hash.new { |h, name| h[name] = NOMS::Light::Player.new(name) }
        @teams = Hash.new { |h, k| h[k] = TeamData.new(k) }
        @players = Hash.new { |h, k| h[k] = PlayerData.new(k) }
        @winner = nil
        @current_round = Round.new
        @number_rounds = 0
        @player_lefts = []
        @achievements = [
          NOMS::Achievement::PerfectMatch.new(self),
          NOMS::Achievement::MinimumMatch.new(self),
          NOMS::Achievement::MatchWinnerRoundLoser.new(self),
          NOMS::Achievement::MatchWinnerRoundWinner.new(self),
          NOMS::Achievement::Domination.new(self)
        ]
      end

      include Events
      include Util::RenameHelper
      
      attr_reader :teams
      attr_reader :players
      attr_reader :winner
      attr_reader :current_round
      attr_reader :number_rounds

      def multiplayer?
        @players.length >= 2 # && (@teams.length >= 2 || (completed? && !@teams.has_key?(@winner.team)))
      end

      def completed?
        !!@winner
      end

      def inspect
        s = "== Match\n"
        s += "Winner: #{self.completed? ? @winner.team : "No winner"}\n"
        s += "Number rounds: #{@number_rounds}\n"
        s += "Players: " + @players.keys.join(", ") + "\n"
        s += "Teams: " + @teams.keys.join(", ") + "\n"
        s += "Scores:\n"
        @teams.each do |team, team_data|
          t = "#{team} (#{team_data.players.join(", ")})"
          s += "    #{t}: #{team_data.score} #{team_data.rounds_played}\n"
        end
        s
      end
      
      # Stats that should be kept even if the match was not multiplayer
      def merge_informative_stats(real_data)
        @data.each do |player, player_data|
          real_data[player].merge_informative_stats(player_data)
        end
      end
      
      # Stats the should only be kept if the match was multiplayer
      def merge_stats(real_data, match_number)
        @players.each do |player, player_data|
          team_score = player_data.team.any? ? @teams[player_data.team].score : 0
          real_data[player].update_matches(player_data.score, team_score, completed?())
        end
        
        won_by_amount = match_won_by_amount()
        win_players(@winner).each do |player|
          number_matches_won = real_data[player].add_match_win(match_number, @number_rounds, won_by_amount)
          
          @achievements.each do |achievement|
            real_data[player].add_achievement(match_number, number_matches_won, achievement)
          end
        end
        
        @data.each do |player, player_data|
          real_data[player].merge(player_data)
        end
        
        # To reduce memory usage when keeping a list of all matches
        @data = nil
        @achievements = nil
      end
      
      def win_players(win)
        return [] if win.nil?
        return [] if win.team.nil? || win.team.empty?
        return win.players if win.players && win.players.any?
        return @teams[win.team].players
      end
      
      def to_xml(root_node)
        return if !clean_match_parse?()
        
        root_node.match(:number_rounds => @number_rounds) { |match_node|
          current_place = 1
          previous_team = nil
          teams_sorted_by_score(true).each do |team|
            next if team.players.empty?
            
            match_node.team(:name => team.name, :place => current_place) { |team_node|
              team.players.each do |player|
                team_node.player(player, :rounds_played => @players[player].rounds_played)
              end
            }
            
            current_place += 1 if previous_team.nil? || previous_team.score != team.score
            previous_team = team
          end
        }
      end
      
      protected
      
      def process_round_scores
        team_scores = if @current_round.round_score_team_feature?
          @current_round.round_scores_team.dup
        else
          # We need to guess what the team round score was
          teams = Hash.new { |h, k| h[k] = Struct.new(:score, :team).new(0, k) }
          @current_round.round_scores.each do |player_score|
            next if player_score.spectator?
            teams[player_score.team].score += player_score.score
          end
          teams.values
        end

        team_scores.each do |score|
          @teams[score.team].score += score.score
          @teams[score.team].rounds_played += 1
        end

        # Do player round scores. No magic guessing needed.
        @current_round.round_scores.each do |score|
          check_player_team(score.player, score.team)
          
          next if score.spectator?
          
          data = @players[score.player]
          
          data.rounds_played += 1
          data.score += score.score
        end

        @player_lefts.each do |left|
          check_team_dissolve(left.player)
        end

        @teams.delete_if do |team, team_data|
          team_data.players.empty?
        end

        # @teams.delete_if do |team, team_data|
        #   !team_data.players.all? { |p| @players.has_key?(p) }
        # end
        
        record_stats(team_scores)
      end
      
      def record_stats(team_scores)
        team_scores.each do |team_score|
          opponents_score = (team_scores - [team_score]).inject(0) { |memo, t| memo + t.score }
          @teams[team_score.team].players.each do |player|
            @data[player].update_team_round_score(team_score.score, opponents_score)
          end
        end
        
        @current_round.round_scores.each do |score|
          @data[score.player].update_round_score(score.score, score.team && score.team.any?)
        end
        
        @current_round.death_suicides.each do |suicide|
          @data[suicide.player].update_suicide
        end
        
        @current_round.death_frags.each do |frag|
          @data[frag.prey].update_prey(frag.hunter)
          @data[frag.hunter].update_hunter(frag.prey)
        end
        
        @current_round.death_teamkills.each do |teamkill|
          @data[teamkill.prey].update_teamkilled
          @data[teamkill.hunter].update_teamkiller
        end
        
        @current_round.basezones_conquered.each do |zone|
          @teams[zone.team].players.each do |player|
            @data[player].update_basezone_conquered
          end
          
          zone.conquerers.each do |conquerer|
            if @players[conquerer.player].team == zone.team
              @data[conquerer.player].update_basezone_conqueror_defender
            else
              @data[conquerer.player].update_basezone_conqueror
            end
          end
        end
        
        @current_round.sacrifices.each do |sacrifice|
          @data[sacrifice.hole_user].update_hole_used
          @data[sacrifice.hole_maker].update_hole_made
          @data[sacrifice.enemy_holed].update_holed
        end
        
        @current_round.cycles_created.each do |cycle|
          @data[cycle.player].update_cycle_created
        end
        
        @current_round.targetzone_player_enters.each do |zone|
          @data[zone.player].update_target_zone
        end
        
        win_players(@current_round.round_winner).each do |player|
          @data[player].add_round_win
        end
      end

      def check_player_team(player, team)
        player_data = @players[player]

        return if player_data.team == team

        remove_player_from_unknown_team(player)
        player_data.rounds_played = 0
        player_data.team = team
        @teams[team].players << player if team.any?
      end

      def remove_player_from_unknown_team(player)
        @teams.each_value do |team|
          team.players.delete(player)
        end
      end

      def check_team_dissolve(player)
        any_scores = @current_round.round_scores.any? { |score| score.player == player }
        if !any_scores
          player_data = @players[player]
          if player_data.team.any?
            team_data = @teams[player_data.team]
            team_data.players.delete(player)
            @teams.delete(player_data.team) if team_data.players.empty?
          end
        end
      end
      
      # We can not accurately parse all matches from ladderlog.txt. For example, in a
      # team game if a team name changes, then the parsed match result may be different
      # than what actually happened. When the outcome of the match is important, you 
      # should ensure at a basic level that this is a clean match.
      def clean_match_parse?
        return false if !completed?() || !@teams.has_key?(@winner.team)
        number_teams_with_players = @teams.values.inject(0) { |memo, team| memo + (team.players.any? ? 1 : 0) }
        winner_has_players = @teams[@winner.team].players.any?
        return number_teams_with_players > 1 && winner_has_players
      end
      
      def teams_sorted_by_score(ensure_winner_first=false)
        sorted_teams = @teams.values.sort { |a, b| b.score <=> a.score }
        
        if ensure_winner_first && sorted_teams[0].name != @winner.team
          winning_team = sorted_teams.find { |team| team.name == @winner.team }
          sorted_teams.delete(winning_team)
          sorted_teams.unshift(winning_team)
        end
        
        return sorted_teams
      end
            
      def match_won_by_amount
        return 0 if !clean_match_parse?()
        sorted_teams = teams_sorted_by_score()
        return 0 if sorted_teams[0].name != @winner.team
        return sorted_teams[0].score - sorted_teams[1].score
      end

    end

    class Parser
      def initialize(stream, last_new_match_pos)
        @stream = stream
        @last_new_match_pos = last_new_match_pos
        @current_new_match_pos = 0
      end
      
      attr_reader :current_new_match_pos

      def each_match
        match = Match.new
        
        @stream.seek(@last_new_match_pos, IO::SEEK_SET)
        @stream.each do |line|
          base = Action::Base.new(line)
          
          case base.name
          when "NEW_MATCH"
            @current_new_match_pos = @stream.pos
            match.match_end
            yield(match)
            match = Match.new
            match.new_match
          when "GAME_END"
            match.match_end
            yield(match)
            match = Match.new
          when "PLAYER_RENAMED"
            match.player_renamed(Action::PlayerRenamed.new(base, HAS_IP_ADDRESSES, STY_CT))
          when "PLAYER_ENTERED"
            match.player_entered(Action::PlayerEntered.new(base, HAS_IP_ADDRESSES))
          when "NEW_ROUND"
            match.new_round
          when "ROUND_SCORE"
            match.round_score(Action::RoundScore.new(base))
          when "ROUND_SCORE_TEAM"
            match.round_score_team(Action::RoundScoreTeam.new(base))
          when "DEATH_SUICIDE"
            match.death_suicide(Action::DeathSuicide.new(base))
          when "DEATH_FRAG"
            match.death_frag(Action::DeathFrag.new(base))
          when "DEATH_TEAMKILL"
            match.death_teamkill(Action::DeathTeamkill.new(base))
          when "MATCH_WINNER"
            match.match_winner(Action::MatchWinner.new(base))
          when "ROUND_WINNER"
            match.round_winner(Action::RoundWinner.new(base))
          when "PLAYER_LEFT"
            match.player_left(Action::PlayerLeft.new(base))
          when "BASEZONE_CONQUERED"
            match.basezone_conquered(Action::BasezoneConquered.new(base))
          when "BASEZONE_CONQUERER"
            match.basezone_conquerer(Action::BasezoneConquerer.new(base))
          when "SACRIFICE"
            match.sacrifice(Action::Sacrifice.new(base))
          when "CYCLE_CREATED"
            match.cycle_created(Action::CycleCreated.new(base))
          when "TARGETZONE_PLAYER_ENTER"
            match.targetzone_player_enter(Action::TargetzonePlayerEnter.new(base))
          end
        end
      end
      
      def all_matches
        enum_for(:each_match).to_a
      end
    end
    
    class PlayerInformationParser
      SCREEN_NAMES_LIMIT = 5
      LINES_LIMIT = 100_000
      
      def initialize(stream, data, last_new_match_pos)
        @stream = stream
        @data = data
        @last_new_match_pos = last_new_match_pos
        @screen_names_needed = Hash.new { |h, player| h[player] = SCREEN_NAMES_LIMIT }
        initialize_names_needed()
      end
      
      def parse
        @stream.seek(@last_new_match_pos, IO::SEEK_SET)
        @stream.enum_for(:reverse_each, false).each_with_index do |line, index|
          break if index >= LINES_LIMIT || finished_parsing?()
          
          base = Action::Base.new(line)
          
          case base.name
          when "PLAYER_RENAMED"
            action = Action::PlayerRenamed.new(base, HAS_IP_ADDRESSES, STY_CT)
            @data[action.new_name].update_screen_name(action.screen_name, action.ip_address) if interesting?(action.new_name)
          when "PLAYER_ENTERED"
            action = Action::PlayerEntered.new(base, HAS_IP_ADDRESSES)
            @data[action.player].update_screen_name(action.screen_name, action.ip_address) if interesting?(action.player)
          end
        end
      end
      
      protected
      
      def interesting?(player)
        @screen_names_needed.has_key?(player) && @screen_names_needed[player] > 0
      end
      
      def finished_parsing?
        @screen_names_needed.values.all? { |n| n <= 0 }
      end
      
      def initialize_names_needed
        @data.each do |player, player_data|
          @screen_names_needed[player] -= player_data.number_screen_names_used
        end
      end
    end
    
    def self.should_process_ladderlog?(stream)
      last_size = last_ladderlog_number_bytes()
      stream.seek(0, IO::SEEK_END)
      current_size = stream.pos
      stream.seek(0, IO::SEEK_CUR)
      last_size.zero? || current_size != last_size
    end
    
    def self.process_ladderlog(stream, data)
      matches_parsed = 0
      match_number = current_match_number()
      ladderlog = Ladderlog::Parser.new(stream, current_new_match_pos())
      matches = []
      
      ladderlog.each_match do |match|
        if NOMS.options[:matches_limit] > 0 && matches_parsed >= NOMS.options[:matches_limit]
          break
        end
        
        match.merge_informative_stats(data)
        next if !match.multiplayer?
        match.merge_stats(data, match_number + 1)
        if match.completed?
          matches << match if NOMS.config.armaskill.generate
          match_number += 1
          matches_parsed += 1
        end
      end
      
      return [matches, ladderlog, match_number]
    end
    
    def self.process_player_information(stream, data)
      parser = Ladderlog::PlayerInformationParser.new(stream, data, current_new_match_pos())
      parser.parse
    end
    
    protected
    
    def self.current_new_match_pos
      return NOMS::DB::Meta.first.current_new_match_pos if NOMS.options[:partial_update]
      return 0
    end
    
    def self.current_match_number
      return NOMS::DB::Meta.first.current_match_number if NOMS.options[:partial_update]
      return 0
    end
    
    def self.last_ladderlog_number_bytes
      return NOMS::DB::Meta.first.ladderlog_number_bytes if NOMS.options[:partial_update]
      return 0
    end
  end
end
