# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2010  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

module NOMS
  module Ladderlog
    module Action

      class Base
        def initialize(line)
          @name, @line = parse_action_name(line)
        end

        attr_reader :name
        attr_reader :line

        def parse_regular_tokens
          @line.split(/ /)
        end

        def parse_special_token(after)
          line =~ /(?:\S+ ){#{after}}(.*)\n?/
          $1
        end

        protected

        def parse_action_name(line)
          line =~ /(\S+) ?(.*)\n?/
          [$1, $2]
        end
      end

      class BaseAction
        def initialize(base)
          @base = base
          @tokens = base.parse_regular_tokens()
          parse()
        end
        attr_reader(:base)
      end

      class RoundScore < BaseAction
        def parse
          @score = @tokens[0].to_i
          @player = @tokens[1]
          @team = @tokens[2] || ""
        end
        attr_reader(:score, :player, :team)
        
        def spectator?
          @team.nil? || @team.empty?
        end
      end

      class RoundScoreTeam < BaseAction
        def parse
          @score = @tokens[0].to_i
          @team = @tokens[1]
        end
        attr_reader(:score, :team)
      end

      class DeathSuicide < BaseAction
        def parse
          @player = @tokens[0]
        end
        attr_reader(:player)
      end

      class DeathFrag < BaseAction
        def parse
          @prey = @tokens[0]
          @hunter = @tokens[1]
        end
        attr_reader(:prey, :hunter)
      end

      class DeathTeamkill < BaseAction
        def parse
          @prey = @tokens[0]
          @hunter = @tokens[1]
        end
        attr_reader(:prey, :hunter)
      end

      class MatchWinner < BaseAction
        def parse
          @team = @tokens[0]
          @players = @tokens[1..-1]
        end
        attr_reader(:team, :players)
      end

      class RoundWinner < MatchWinner
      end

      class PlayerEntered < BaseAction
        def initialize(base, has_ip_addresses)
          @has_ip_addresses = has_ip_addresses
          super(base)
        end

        def parse
          @player = @tokens[0]
          @ip_address = @has_ip_addresses ? @tokens[1] : ""
          @screen_name = @base.parse_special_token(@has_ip_addresses ? 2 : 1) || ""
        end
        attr_reader(:player, :ip_address, :screen_name)
      end

      class PlayerRenamed < BaseAction
        def initialize(base, has_ip_addresses, is_sty_ct)
          @has_ip_addresses = has_ip_addresses
          @is_sty_ct = is_sty_ct
          super(base)
        end

        def parse
          @old_name = @tokens[0]
          @new_name = @tokens[1]
          @ip_address = @has_ip_addresses ? @tokens[2] : ""

          screen_name_index = @has_ip_addresses ? 3 : 2
          screen_name_index += 1 if @is_sty_ct

          @screen_name = @base.parse_special_token(screen_name_index) || ""
        end

        attr_reader(:old_name, :new_name, :ip_address, :screen_name)

        def authenticated_rename?
          !!@new_name.index("@")
        end
      end

      class PlayerLeft < BaseAction
        def parse
          @player = @tokens[0]
        end
        attr_reader(:player)
      end

      class BasezoneConquered < BaseAction
        def initialize(base)
          @conquerers = []
          super(base)
        end
        
        def parse
          @team = @tokens[0]
        end
        
        attr_reader(:team, :conquerers)
      end

      class BasezoneConquerer < BaseAction
        def parse
          @player = @tokens[0]
        end
        attr_reader(:player)
      end

      class Sacrifice < BaseAction
        def parse
          @hole_user = @tokens[0]
          @hole_maker = @tokens[1]
          @enemy_holed = @tokens[2]
        end
        attr_reader(:hole_user, :hole_maker, :enemy_holed)
      end

      class CycleCreated < BaseAction
        def parse
          @player = @tokens[0]
        end
        attr_reader(:player)
      end

      class TargetzonePlayerEnter < BaseAction
        def parse
          @player = @tokens[4]
        end
        attr_reader(:player)
      end
    end
    
  end
end