# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "rubygems"
require "sequel"
require "sequel/extensions/pagination"
require "set"

class NOMS::DB::Player
  class NotFound < StandardError; end
  
  PLAYERS_PER_PAGE = 100
  
  one_to_many :match_wins
  one_to_many :round_wins
  one_to_many :perfect_match_achievements
  one_to_many :minimum_match_achievements
  one_to_many :match_winner_round_loser_achievements
  one_to_many :match_winner_round_winner_achievements
  one_to_many :domination_achievements
  
  def self.filter_data(dataset)
    if NOMS.config.sanity_checks.filter_no_rounds_played
      return dataset.filter(:rounds_played > 0)
    end

    return dataset
  end
  
  def self.number_players
    self.filter_data(NOMS::DB::Player).count
  end
  
  def self.stats_data(tier)
    self.filter_data(self.order_by(:won_matches, :won_rounds, :won_rounds_ladderlog, :round_points_total)).reverse.paginate(tier, PLAYERS_PER_PAGE).eager(
      :match_wins
    )
  end
  
  def self.ratings_data
    self.filter_data(self.order_by(:rating)).limit(PLAYERS_PER_PAGE).reverse.eager(:match_wins)
  end
  
  def self.search_players(name, limit=10)
    field = name.include?("@") ? :name : :screen_name
    
    # Try for exact matches
    players = filter(field => name)
    return players if players.count == 1
    
    # OK, no exact matches
    filter(field.like(name)).limit(limit)
  end
  
  def self.find_by_name(name)
    player = self.find(:name => name)
    if !player
      raise NotFound.new("Can't find player named “#{name}”")
    end
    player
  end
  
  def self.formatted_match_wins(name)
    player = find_by_name(name)
    match_wins = player.match_wins_dataset.all
    
    if match_wins.empty?
      raise StandardError.new("The player “#{name}” has no match history records")
    end
    
    match_wins.map { |win|
      [win.global_match_number, win.player_match_win_number]
    }
  end
  
  def self.countries
    query = "SELECT DISTINCT country_code from players"
    query += " WHERE `rounds_played` > 0" if NOMS.config.sanity_checks.filter_no_rounds_played
    NOMS.current_db[query].map { |c| c[:country_code] }.reject { |c| c.empty? }
  end
  
  def self.countries_simple_counts
    countries = self.countries
    return {} if countries.empty?
    queries = countries.inject([]) { |memo, country| 
      
      query = "(SELECT COUNT(*) FROM `players` WHERE `country_code` = '#{country}'"
      query += " AND `rounds_played` > 0" if NOMS.config.sanity_checks.filter_no_rounds_played
      query += ") as '#{country}'"
      memo <<  query
    }.join(", ")
    
    self.filter_data(NOMS.current_db["SELECT #{queries}"]).first.inject({}) { |memo, (country, count)|
      memo[country.to_s] = count.to_i
      memo
    }
  end
  
  def self.number_players_from_country(a2)
    filter(:country_code => a2).count
  end
  
  def merge_light(other)
    last_match_win = self.match_wins_dataset.order_by(:player_match_win_number).last
    if last_match_win
      win_number = last_match_win.player_match_win_number
      
      other.match_wins.each do |win|
        win.player_match_win_number += win_number
      end
      
      other.achievements.each_value do |achievements|
        achievements.each { |a| a.player_match_win_number += win_number }
      end
    end
    
    merge(other)    
  end
  
  alias_method(:old_merge, :merge)
  def merge(other)
    old_merge(other)
    
    if other.screen_name.any? && other.screen_name != self.screen_name
      self.screen_name = other.screen_name
    end
    
    if other.country_code.any? && other.country_code != self.country_code
      self.country_code = other.country_code
    end
    
    if other.is_armaskill_rated
      self.mu = other.mu
      self.sigma = other.sigma
      self.rating = other.rating
      self.is_armaskill_rated = true
    end
    
    merge_simple_stat(other, :won_matches)
    merge_simple_stat(other, :won_rounds)
    merge_simple_stat(other, :ladder)
  end
    
  def teamkilled_average
    check_number(1, self.teamkilled_total.to_f / self.rounds_played)
  end
  
  def teamkiller_average
    check_number(1, self.teamkiller_total.to_f / self.rounds_played)
  end
  
  def holes_used_average
    check_number(0, self.holes_used_total.to_f / self.rounds_played)
  end
  
  def holes_made_average
    check_number(1, self.holes_made_total.to_f / self.rounds_played)
  end
  
  def holed_total_average
    check_number(1, self.holed_total.to_f / self.rounds_played)
  end
  
  def basezone_conquered_avg
    check_number(1, self.basezone_conquered.to_f / self.rounds_played)
  end
  
  def basezone_conqueror_defender_avg
    check_number(0, self.basezone_conqueror_defender.to_f / self.rounds_played)
  end
  
  def basezone_conqueror_avg
    check_number(0, self.basezone_conqueror.to_f / self.rounds_played)
  end
  
  def average_match_length
    check_number(1.0 / 0, self.rounds_played.to_f / self.matches_played)
  end
  
  def won_matches_for_display
    if !NOMS.config.sanity_checks.has_simple_stats
      return self.match_wins.length
    end
    
    self.won_matches
  end
  
  def won_rounds_for_display
    if !NOMS.config.sanity_checks.has_simple_stats
      return self.won_rounds_ladderlog
    end
    
    self.won_rounds
  end
  
  def concrete_won_rounds
    if !NOMS.config.sanity_checks.ladderlog_complete
      return self.won_rounds_ladderlog
    end
    
    self.won_rounds
  end
  
  def concrete_won_matches
    if !NOMS.config.sanity_checks.ladderlog_complete
      # return self.match_wins_dataset.count
      return self.match_wins.length
    end
    
    self.won_matches
  end
    
  def matches_percent_won_global
    self.concrete_won_matches.to_f / (self.rounds_played / NOMS::DB::MatchWin.global_average_match_length)
  end
  
  def matches_percent_won
    check_number(0, self.concrete_won_matches.to_f / self.matches_played)
  end
  
  def matches_average_number_rounds
    # self.match_wins_dataset.avg(:number_rounds).to_f
    check_number(1.0 / 0, avg_by(self.match_wins) { |win| win.number_rounds })
  end
  
  def matches_min_number_rounds
    # self.match_wins_dataset.min(:number_rounds).to_i
    min_by(self.match_wins, 0) { |win| win.number_rounds }
  end
  
  def matches_max_number_rounds
    # self.match_wins_dataset.max(:number_rounds).to_i
    max_by(self.match_wins, 0) { |win| win.number_rounds }
  end
  
  def rounds_percent_won
    check_number(0, self.concrete_won_rounds.to_f / self.rounds_played)
  end
  
  def points_average
    check_number(0, self.round_points_total.to_f / self.rounds_played)
  end
  
  def team_points_average
    check_number(0, self.team_round_points_total.to_f / self.rounds_played)
  end
  
  def opponents_points_average
    check_number(1.0 / 0, self.opponents_round_points_total.to_f / self.rounds_played)
  end
    
  def team_points_diff
    self.team_points_average - self.opponents_points_average
  end
  
  def points_percent_zero
    check_number(1, self.rounds_with_zero_points.to_f / self.rounds_played)
  end
  
  def core_dumps_percent_dumped
    check_number(1, self.prey_total.to_f / self.rounds_played)
  end
  
  def core_dumps_hunter_average
    check_number(0, self.hunter_total.to_f / self.rounds_played)
  end
  
  def deaths_total
    self.suicides_total + self.prey_total + self.teamkilled_total
  end
  
  def kill_death_ratio
    check_number(0, self.hunter_total.to_f / self.deaths_total)
  end
    
  def suicides_percent_suicided
    check_number(1, self.suicides_total.to_f / self.rounds_played)
  end
  
  def won_by_avg    
    # self.match_wins_dataset.avg(:won_by_amount).to_f
    check_number(0, avg_by(self.match_wins) { |win| win.won_by_amount })
  end
  
  def won_by_max
    # self.match_wins_dataset.max(:won_by_amount).to_i
    max_by(self.match_wins, 0) { |win| win.won_by_amount }
  end
  
  def target_zone_completion_percent
    check_number(0, self.target_zones_entered.to_f / self.cycles_created)
  end
  
  def target_zone_avg
    check_number(0, self.target_zones_entered.to_f / self.rounds_played)
  end
  
  def display_name
    return self.name if !self.screen_name.any?
    return "#{self.screen_name} (#{self.name})" if self.name.index("@")
    return self.screen_name
  end
  
  def won_matches_rank
    rank(:won_matches)
  end
  
  def won_rounds_rank
    rank(:won_rounds)
  end
  
  def ladder_rank
    rank(:ladder)
  end
  
  private
  def rank(column)
    self.class.filter(column > self.method(column).call).count + 1
  end
  
  def check_number(replacement, n)
    return replacement if n.infinite? || n.nan?
    n
  end
  
  def sum_by(xs)
    xs.inject(0) { |memo, x| memo + yield(x) }
  end
  
  def avg_by(xs, &blk)
    sum_by(xs, &blk).to_f / xs.length
  end
  
  def min_by(xs, default=nil, &blk)
    return default if xs.empty? && default
    m = xs.min { |a, b| yield(a) <=> yield(b) }
    return yield(m)
  end
  
  def max_by(xs, default=nil, &blk)
    return default if xs.empty? && default
    m = xs.max { |a, b| yield(a) <=> yield(b) }
    return yield(m)
  end
  
  def merge_simple_stat(other, column)
    other_value = other.method(column).call
    return if other_value.zero?
    self.method("#{column}=").call(other_value)
  end
end

module NOMS::Distinct
  def distinct_players_count
    NOMS.current_db["SELECT DISTINCT player_id from ?", self.table_name].count
  end
  
  def distinct_matches_count
    NOMS.current_db["SELECT DISTINCT global_match_number from ?", self.table_name].count
  end
end

class NOMS::DB::RoundWin
  many_to_one :player
end

class NOMS::DB::MatchWin
  many_to_one :player
  extend NOMS::Distinct
  
  def self.global_average_match_length
    if @average_match_length
      return @average_match_length
    end
    
    @average_match_length = self.avg(:number_rounds).to_f
  end  
end

class NOMS::DB::PerfectMatchAchievement
  many_to_one :player
  extend NOMS::Distinct
end

class NOMS::DB::MinimumMatchAchievement
  many_to_one :player
  extend NOMS::Distinct
end

class NOMS::DB::MatchWinnerRoundLoserAchievement
  many_to_one :player
  extend NOMS::Distinct
end

class NOMS::DB::MatchWinnerRoundWinnerAchievement
  many_to_one :player
  extend NOMS::Distinct
end

class NOMS::DB::DominationAchievement
  many_to_one :player
  extend NOMS::Distinct
end
