# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

module NOMS
  
  class ModelGenerator
    def initialize(klass, table_name, should_create_tables, should_create_db)
      @generator = should_create_db ? Sequel::Schema::Generator.new(NOMS.current_db) { primary_key(:id) } : Util::MethodNotMissing.new
      @attributes = []
      yield(self) if block_given?
      _define(klass, table_name, should_create_tables, should_create_db)
    end
  
    attr_reader :generator

    def _define(klass_name, table_name, should_create_tables, should_create_db)
      light_class = Class.new
      heavy_class = Class.new(Sequel::Model) if should_create_db
    
      NOMS::Light.const_set(klass_name, light_class)
      
      if should_create_db
        NOMS::DB.const_set(klass_name, heavy_class)
        heavy_class.const_set(:ATTRIBUTES, @attributes)
      end
      
      @attributes.each do |(name, kind, opts)|
        light_class.__send__(:attr_accessor, name)
        @generator.column(name, kind, opts)
      end
    
      if should_create_tables
        NOMS.current_db.create_table(table_name, :generator => @generator)
      end
      
      if should_create_db
        heavy_class.set_dataset(table_name)
        heavy_class.set_primary_key(:id)
      end
      
      merge_method = lambda { |other|
        heavy_class()::ATTRIBUTES.each do |(name, type, opts, merge_type)|
          setter = self.method(:"#{name}=")
          value = self.method(name).call
          other_value = other.method(name).call
          
          case merge_type
          when :default
            setter.call(other_value) if other_value != value
          when :sum
            setter.call(value + other_value) if other_value.nonzero?
          when :max
            setter.call(other_value) if other_value > value
          when :no_merge
          end
        end
      }
      
      light_class.class_eval do
        if should_create_db
          define_method(:merge, &merge_method)
          define_method(:heavy_class) do
            NOMS::DB.const_get(self.class.to_s.split("::").last)
          end
        
          define_method(:to_heavy) do
            heavy = heavy_class()
            values = heavy::ATTRIBUTES.inject({}) { |memo, (name, type, opts, merge_type)|
              memo[name] = method(name).call
              memo
            }
            return heavy.create(values)
          end
        else
          define_method(:to_heavy) { }
        end
      end
            
      if should_create_db
        heavy_class.class_eval do
          define_method(:heavy_class) {
            return self.class
          }
          define_method(:merge, &merge_method)
        end
      end
    end
  
    def attribute(name, kind, opts={})
      @attributes << [name, kind, opts, :default]
    end
    
    def sum_attribute(name, kind, opts={})
      @attributes << [name, kind, opts, :sum]
    end
    
    def max_attribute(name, kind, opts={})
      @attributes << [name, kind, opts, :max]
    end
    
    def no_merge_attribute(name, kind, opts={})
      @attributes << [name, kind, opts, :no_merge]
    end
  end

end
