# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "iconv"

class NOMS::Light::Player
  def initialize(name)
    self.name = name
    self.round_points_highscore = 0
    self.round_points_total = 0
    self.rounds_played = 0
    self.match_points_highscore = 0
    self.matches_played = 0
    self.suicides_total = 0
    self.won_matches = 0
    self.won_rounds = 0
    self.won_rounds_ladderlog = 0
    self.ladder = 0
    self.prey_total = 0
    self.hunter_total = 0
    self.rounds_with_zero_points = 0
    self.mu = 0.0
    self.sigma = 0.0
    self.rating = 0.0
    self.is_armaskill_rated = false
    self.basezone_conquered = 0
    self.basezone_conqueror_defender = 0
    self.basezone_conqueror = 0
    self.holes_used_total = 0
    self.holes_made_total = 0
    self.holed_total = 0
    self.teamkilled_total = 0
    self.teamkiller_total = 0
    self.team_round_points_total = 0
    self.opponents_round_points_total = 0
    self.team_round_points_highscore = 0
    self.team_match_points_highscore = 0
    self.match_win_streak_strict = 0
    self.cycles_created = 0
    self.target_zones_entered = 0
    self.screen_name = ""
    self.country_code = ""
    
    @match_wins = []
    @round_wins = []
    @current_match_win_streak_strict = 0
    @achievements = Hash.new { |h, k| h[k] = [] }
    @screen_names = Hash.new { |h, k| h[k] = 0 }
    @ip_addresses = Hash.new { |h, k| h[k] = 0 }
  end
  
  attr_reader :match_wins
  attr_reader :achievements
  attr_reader :screen_names
  attr_reader :ip_addresses
  
  def prepare_merge
    most_used_screen_name = @screen_names.max { |a, b| a[1] <=> b[1] }
    if most_used_screen_name
      name = most_used_screen_name[0]
      name = Iconv.conv("UTF8", "ISO-8859-1", name) if !NOMS.config.sanity_checks.ladderlog_utf8
      self.screen_name = name
    end
    self.country_code = most_used_country_code() if NOMS::Choropleth.available?
  end
  
  alias_method :orig_to_heavy, :to_heavy
  def to_heavy(player=nil)
    player ||= orig_to_heavy()
    @match_wins.each { |win| player.add_match_win(win.to_heavy) }
    @round_wins.each { |win| player.add_round_win(win.to_heavy) }
    @achievements.each do |add_method, achievements|
      achievements.each { |win| player.method(add_method).call(win.to_heavy) }
    end
    return player
  end
  
  def merge_informative_stats(other)
    other.screen_names.each do |screen_name, usages|
      @screen_names[screen_name] += usages
    end
    other.ip_addresses.each do |ip_address, usages|
      @ip_addresses[ip_address] += usages
    end
  end
  
  def update_screen_name(screen_name, ip_address)
    @screen_names[screen_name] += 1 if screen_name.any?
    @ip_addresses[ip_address] += 1 if ip_address && ip_address.any?
  end
  
  def number_screen_names_used
    @screen_names.values.inject(0) { |memo, n| memo + n }
  end
  
  def add_match_win(global_id, number_rounds, won_by_amount)
    # TODO abstract this out so it can be used for round win streaks
    # TODO and non-strict win streak behavior
    if @current_match_win_streak_strict > 0 || @match_wins.empty?
      if @match_wins.empty?
        @current_match_win_streak_strict += 1
      else
        last_win = @match_wins[-1]
        if last_win.global_match_number + 1 == global_id
          @current_match_win_streak_strict += 1
        else
          real_streak = @current_match_win_streak_strict + 1
          self.match_win_streak_strict = real_streak if real_streak > self.match_win_streak_strict
          @current_match_win_streak_strict = 0
        end
      end
    end
    
    win = NOMS::Light::MatchWin.new
    win.global_match_number = global_id
    win.player_match_win_number = @match_wins.length + 1
    win.number_rounds = number_rounds
    win.won_by_amount = won_by_amount
    @match_wins << win
    return win.player_match_win_number
  end
  
  def add_round_win(global_id_ignored=0)
    self.won_rounds_ladderlog += 1
    
    # win = NOMS::Light::RoundWin.new
    # win.global_round_number = global_id
    # win.player_round_win_number = @round_wins.length + 1
    # @round_wins << win
  end
  
  def add_achievement(match_number, number_matches_won, achievement)
    if achievement.accomplished?(self.name)
      @achievements[achievement.add_method] << achievement.to_light(match_number, number_matches_won)
    end
  end
  
  def update_round_score(score, is_playing)
    # This player is a spectator
    return if !is_playing
    
    # ROUND_SCORE for spectators is broken on sty+ct. See bug <https://bugs.launchpad.net/armagetronad/+bug/537147>
    if (NOMS::Ladderlog::MAX_ROUND_SCORE <= 0 || score <= NOMS::Ladderlog::MAX_ROUND_SCORE) && \
        (NOMS::Ladderlog::MIN_ROUND_SCORE == "default" || score >= NOMS::Ladderlog::MIN_ROUND_SCORE)
      self.round_points_total += score
      self.round_points_highscore = score if score > self.round_points_highscore
    end
    
    self.rounds_with_zero_points += 1 if score == 0
    self.rounds_played += 1
  end
  
  def update_team_round_score(team_score, opponents_score)
    self.team_round_points_total += team_score
    self.opponents_round_points_total += opponents_score
    self.team_round_points_highscore = team_score if team_score > self.team_round_points_highscore
  end
  
  def update_suicide
    self.suicides_total += 1
  end
  
  def update_prey(hunter)
    self.prey_total += 1
  end
  
  def update_hunter(prey)
    self.hunter_total += 1
  end
  
  def update_teamkilled
    self.teamkilled_total += 1
  end
  
  def update_teamkiller
    self.teamkiller_total += 1
  end
  
  def update_basezone_conquered
    self.basezone_conquered += 1
  end
  
  def update_basezone_conqueror_defender
    self.basezone_conqueror_defender += 1
  end
  
  def update_basezone_conqueror
    self.basezone_conqueror += 1
  end
  
  def update_hole_used
    self.holes_used_total += 1
  end
  
  def update_hole_made
    self.holes_made_total += 1
  end
  
  def update_holed
    self.holed_total += 1
  end
  
  def update_cycle_created
    self.cycles_created += 1
  end
  
  def update_target_zone
    self.target_zones_entered += 1
  end
  
  def update_matches(match_score, team_match_score, match_completed)
    self.matches_played += 1 if match_completed
    self.match_points_highscore = match_score if match_score > self.match_points_highscore
    self.team_match_points_highscore = team_match_score if team_match_score > self.team_match_points_highscore
  end
  
  # TODO implement
  def _win_streaks(history)
    # ...
  end

  def match_win_streaks
    _win_streaks(@won_matches_history)
  end

  def round_win_streaks
    _win_streaks(@won_rounds_history)
  end
  
  protected
  
  def most_used_country_code
    country_codes = Hash.new { |h, k| h[k] = 0 }
    @ip_addresses.each do |ip, usages|
      entry = NOMS::Choropleth::Entry.lookup_ip(ip)
      country_codes[entry.country_code] += usages if entry
    end
    country_code = country_codes.max { |a, b| a[1] <=> b[1] }
    return country_code ? country_code[0] : ""
  end
end

class NOMS::Light::Meta
  def initialize
    self.current_new_match_pos = 0
    self.current_match_number = 0
    self.ladderlog_number_bytes = 0
    self.old_won_matches_txt = ""
    self.old_won_rounds_txt = ""
    self.old_ladder_txt = ""
  end
end
