# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "fileutils"
require "rubygems"
require "sequel"
require "noms/config"
require "noms/util"
require "logger"
require "net/http"
require "uri"

module NOMS
  module Light; end;
  module DB; end
  
  Server = Struct.new(:config, :db_path) do
    attr_reader :db
    
    def database_modified_time
      if File.exist?(db_path)
        return File.mtime(db_path)
      end
      
      Time.now
    end
    
    def connect_db
      @db = Sequel.connect("sqlite://#{db_path()}")
      if ENV["NOMS_DEBUG"]
        @db.loggers << Logger.new(STDOUT)
      end
    end
  end
  
  class << self
    attr_accessor :servers
    attr_accessor :current_server
    attr_accessor :current_db
    attr_accessor :options
  end
  @servers = {}
  @options = {
    :matches_limit => -1,
    :db_directory => "db",
    :development_mode => false,
    :profile => false,
    :partial_update => false
  }
    
  def self.set_current_db(db)
    @current_db = db
    NOMS::DB.constants.each do |model_class|
      NOMS::DB.const_get(model_class).db = db
    end
  end
  
  def self.set_current_server(server)
    NOMS.config = server.config
    NOMS.set_current_db(server.db)
    NOMS.current_server = server
  end
  
  def self.reconnect_db
    NOMS.current_server.connect_db
    NOMS.set_current_db(NOMS.current_server.db)
  end
    
  module Setup
    REMOTE_VAR_DIR = NOMS::Util.top_path("tmp/var")
    REMOTE_VAR_DIR_FILE = File.join(REMOTE_VAR_DIR, "var.tar.gz")
    
    def self.database_file(should_create_tables, suffix="", ext="sqlite3")
      filename = "#{NOMS.config.short_name}#{suffix}.#{ext}"
      filename += ".temp" if should_create_tables
      return File.join(NOMS.options[:db_directory], filename)
    end
    
    def self.add_config(config_path, should_create_tables=false)
      config = NOMS::Config.new
      config.define_config
      config.from_file(config_path)
      
      NOMS.config = config
      
      # We will move this to the real DB when done
      db_path = database_file(should_create_tables)
      FileUtils.mkdir_p(File.dirname(db_path))
      
      # Old tmp database leftover from a crash
      if should_create_tables && File.exist?(db_path)
        FileUtils.rm(db_path)
      end
      
      server = Server.new(config, db_path)
      server.connect_db()
      NOMS.servers[config.short_name] = server
      NOMS.set_current_server(server)
      return server
    end
    
    def self.finish(partial_update=false)
      if !partial_update
        temp_db = database_file(true)
        real_db = database_file(false)
        FileUtils.rm_rf(real_db)
        FileUtils.mv(temp_db, real_db)
      end
      FileUtils.rm_rf(REMOTE_VAR_DIR)
    end
    
    def self.setup_var_dir
      raise RuntimeError.new("No local var directory specified.") if NOMS.config.var_directory.empty?
      local_var_dir = File.expand_path(NOMS.config.var_directory)
      return local_var_dir if File.exist?(local_var_dir)
      raise RuntimeError.new("No var directory specified.") if NOMS.config.var_directory_uri.empty?
      puts "Fetching #{NOMS.config.var_directory_uri}"
      self.fetch_remote_var(NOMS.config.var_directory_uri)
      puts "Extracting #{REMOTE_VAR_DIR_FILE}"
      system("tar xfz #{NOMS::Util.e_sh REMOTE_VAR_DIR_FILE} -C #{NOMS::Util.e_sh REMOTE_VAR_DIR}")
      REMOTE_VAR_DIR
    end
    
    def self.generate_models(should_create_tables, should_create_db=true)
      ModelGenerator.new(:Meta, :meta, should_create_tables, should_create_db) do |m|
        m.attribute(:current_new_match_pos, :integer)
        m.attribute(:current_match_number, :integer)
        m.attribute(:ladderlog_number_bytes, :integer)
        m.attribute(:old_won_matches_txt, :string)
        m.attribute(:old_won_rounds_txt, :string)
        m.attribute(:old_ladder_txt, :string)
      end
      
      ModelGenerator.new(:Player, :players, should_create_tables, should_create_db) do |m|
        m.attribute(:name, :string)
        m.max_attribute(:round_points_highscore, :integer)
        m.sum_attribute(:round_points_total, :integer)
        m.sum_attribute(:rounds_played, :integer)
        m.max_attribute(:match_points_highscore, :integer)
        m.sum_attribute(:matches_played, :integer)
        m.sum_attribute(:suicides_total, :integer)
        m.no_merge_attribute(:won_matches, :integer)
        m.no_merge_attribute(:won_rounds, :integer)
        m.sum_attribute(:won_rounds_ladderlog, :integer)
        m.no_merge_attribute(:ladder, :float)
        m.sum_attribute(:prey_total, :integer)
        m.sum_attribute(:hunter_total, :integer)
        m.sum_attribute(:rounds_with_zero_points, :integer)
        m.no_merge_attribute(:mu, :float)
        m.no_merge_attribute(:sigma, :float)
        m.no_merge_attribute(:rating, :float)
        m.no_merge_attribute(:is_armaskill_rated, :boolean)
        m.sum_attribute(:basezone_conquered, :integer)
        m.sum_attribute(:basezone_conqueror_defender, :integer)
        m.sum_attribute(:basezone_conqueror, :integer)
        m.sum_attribute(:holes_used_total, :integer)
        m.sum_attribute(:holes_made_total, :integer)
        m.sum_attribute(:holed_total, :integer)
        m.sum_attribute(:teamkilled_total, :integer)
        m.sum_attribute(:teamkiller_total, :integer)
        m.sum_attribute(:team_round_points_total, :integer)
        m.sum_attribute(:opponents_round_points_total, :integer)
        m.max_attribute(:team_round_points_highscore, :integer)
        m.max_attribute(:team_match_points_highscore, :integer)
        m.max_attribute(:match_win_streak_strict, :integer)
        m.sum_attribute(:cycles_created, :integer)
        m.sum_attribute(:target_zones_entered, :integer)
        m.no_merge_attribute(:screen_name, :string)
        m.no_merge_attribute(:country_code, :string)
      end

      ModelGenerator.new(:MatchWin, :match_wins, should_create_tables, should_create_db) do |m|
        m.generator.foreign_key(:player_id)

        m.attribute(:global_match_number, :integer)
        m.attribute(:player_match_win_number, :integer)
        m.attribute(:number_rounds, :integer)
        m.attribute(:won_by_amount, :integer)
      end
      
      ModelGenerator.new(:RoundWin, :round_wins, should_create_tables, should_create_db) do |m|
        m.generator.foreign_key(:player_id)

        m.attribute(:global_round_number, :integer)
        m.attribute(:player_round_win_number, :integer)
      end
      
      achievement_match_model = lambda { |m|
        m.generator.foreign_key(:player_id)

        m.attribute(:global_match_number, :integer)
        m.attribute(:player_match_win_number, :integer)
      }
      ModelGenerator.new(:PerfectMatchAchievement, :perfect_match_achievements, should_create_tables, should_create_db, &achievement_match_model)
      ModelGenerator.new(:MinimumMatchAchievement, :minimum_match_achievements, should_create_tables, should_create_db, &achievement_match_model)
      ModelGenerator.new(:MatchWinnerRoundLoserAchievement, :match_winner_round_loser_achievements, should_create_tables, should_create_db, &achievement_match_model)
      ModelGenerator.new(:MatchWinnerRoundWinnerAchievement, :match_winner_round_winner_achievements, should_create_tables, should_create_db, &achievement_match_model)
      ModelGenerator.new(:DominationAchievement, :domination_achievement, should_create_tables, should_create_db, &achievement_match_model)
      
      NOMS::Ladderlog.alias_config_items
      
      require "noms/models/db" if should_create_db
      require "noms/models/light"
      
      if NOMS::Choropleth.available?
        NOMS::Choropleth.connect_db
        NOMS::Choropleth.setup_db(false)
      end
    end
    
    def self.save_meta_ladderlog(ladderlog, match_number, stream)
      meta = NOMS::DB::Meta.first
      stream.seek(0, IO::SEEK_END)
      bytes = stream.tell
      if meta.nil?
        meta = NOMS::Light::Meta.new
        meta.current_new_match_pos = ladderlog.current_new_match_pos
        meta.current_match_number = match_number
        meta.ladderlog_number_bytes = bytes
        meta.to_heavy
      else
        meta.current_new_match_pos = ladderlog.current_new_match_pos
        meta.current_match_number = match_number
        meta.ladderlog_number_bytes = bytes
        meta.save
      end
    end
    
    def self.save(data)
      NOMS.current_db.transaction {
        data.each do |name, light_player|
          light_player.prepare_merge
          
          if NOMS.options[:partial_update]
            heavy_player = NOMS::DB::Player[:name => name]
            
            if heavy_player
              heavy_player.merge_light(light_player)
              light_player.to_heavy(heavy_player)
              heavy_player.save_changes
              next
            end
          end
          
          light_player.to_heavy
        end
      }
    end
    
    private
    def self.fetch_remote_var(uri)
      FileUtils.mkdir_p(REMOTE_VAR_DIR)
      File.open(REMOTE_VAR_DIR_FILE, "wb") do |f|
        f.write(fetch(uri).body)
      end
    end
    
    def self.fetch(uri_str, limit=10)
      # You should choose better exception.
      raise ArgumentError, "HTTP redirect too deep" if limit == 0

      response = Net::HTTP.get_response(URI.parse(uri_str))
      case response
      when Net::HTTPSuccess     then response
      when Net::HTTPRedirection then fetch(response['location'], limit - 1)
      else
        response.error!
      end
    end
  end  
end
