# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "rubygems"
require "sinatra/base"
require "erb"
require "json"
require "sequel"
require "enumerator"
require "rack"
require "rack/cache"

module Enumerable
  def none?(&blk)
    !any?(&blk)
  end
end

class NOMS::WebApp < Sinatra::Base
  class NoServerError < StandardError; end
  
  module Helpers
    include Rack::Utils
    alias_method(:h, :escape_html)
    
    def remove_html(str)
      str.gsub(/<\/?[^>]*>/, "")
    end
    
    def round_to(n, r)
      "%.#{r}f" % n
    end
    
    def round(n)
      round_to(n, 3)
    end
    
    def to_percent(n, round_to=1)
      "#{round_to(n * 100, round_to)}%"
    end
    
    def format_number_human(number, delimiter=",")
      number.to_s.gsub(/(\d)(?=(\d\d\d)+(?!\d))/, "\\1#{delimiter}")
    end
    
    def ordinalize(n)
      suffix = if (11..14).include?(n % 100)
        "th"
      else
        case n % 10
        when 1
          "st"
        when 2
          "nd"
        when 3
          "rd"
        else
          "th"
        end
      end
      "#{n}#{suffix}" 
    end
    
    def kb(bytes)
      "#{bytes / 1024}KB"
    end
    
    def player_name_link(player, html=true)
      path = "/#{NOMS.config.short_name}/player?name=#{Rack::Utils.escape(player.name)}"
      return path if !html
      "<a href=\"#{path}\">#{h player.display_name}</a>"
    end
    
    def stat_signature_link(player, html=true)
      path = "/#{NOMS.config.short_name}/statsig?name=#{Rack::Utils.escape(player.name)}"
      return path if !html
      "<a href=\"#{path}\">Create a stat signature for #{h player.display_name}</a>"
    end
    
    def server_link(server_config, include_group=false, html=true)
      path = "/#{server_config.short_name}"
      return path if !html
      "<a href=\"#{path}\">#{server_name(server_config, include_group)}</a>"
    end
    
    def to_sentence(xs)
      case xs.length
      when 0
        ""
      when 1
        xs[0].to_s
      when 2
        "#{xs[0]} and #{xs[1]}"
      else
        "#{xs[0..-2].join(", ")}, and #{xs[-1]}"
      end
    end
        
    def calculate_common_achievement_stats(stats, number_matches_won_all_time, number_players_all_time)
      stats.keys.each do |key|
        stats[key]["percent_achieved"] = stats[key]["number_distinct_players"].to_f / number_players_all_time
        stats[key]["percent_matches_won"] = stats[key]["number_matches"].to_f / number_matches_won_all_time
      end
    end
    
    def server_name(config=NOMS.config, include_group=true)
      name = h(config.name)
      name = "#{h(config.group)} #{name}" if config.group.any? && include_group
      return name
    end
    
    def server_path
      "/#{NOMS.config.short_name}"
    end
    
    def setup_server(name)
      server = NOMS.servers[name]
      if server.nil?
        raise NoServerError.new("No server configured for “#{name}”.")
      end
      
      NOMS.set_current_server(server)
    end
    
    def stats_tier_link(tier)
      page = tier == 1 ? "stats" : "stats?tier=#{tier}"
      return "<a href=\"#{page}\">Tier #{tier}</a>"
    end
    
    def validate_params_keys(allowed_keys, redirect_page)
      if !params.keys.all? { |p| allowed_keys.any? { |allowed| allowed == p } }
        redirect(redirect_page, 303)
      end
    end
    
    def cached(seconds=90_000) # 25 hours
      response["Cache-Control"] = "max-age=#{seconds}, must-revalidate"
    end
    
    def flag_img_path(player, prefix="/")
      path = "flags/#{player.country_code.downcase}.png"
      return "/#{path}" if File.exist?(NOMS::Util.public_path(path))
      return "/flags/unknown_country.png"
    end
    
    def navbar_class_extra(selected)
      if selected
        return 'class="selected"'
      end
      return ""
    end
    
    def navbar(key)
      page = Hash.new { |h, k| h[k] = false }
      page[key] = true
      erb(:navbar, :locals => {:page => page})
    end
    
    def sort_by_group(servers)
      groups = Hash.new { |h, k| h[k] = [] }
      servers.each do |server|
        groups[server.config.group] << server
      end

      groups = groups.map { |group, grouped_servers|
        [group, NOMS::Util.armasort_by(grouped_servers) { |server| server.config.name }]
      }

      return NOMS::Util.armasort_by(groups) { |(group, servers)| group }
    end
  end
  
  require "noms/web_app/stats_table"
  require "noms/web_app/stat_signature"
  
  class StatsTable
    include Helpers
  end
  
  class StatSignature
    include Helpers
  end
  
  helpers do
    include Helpers
  end
  
  use(
    Rack::Cache,
    :verbose => false,
    :metastore => "file:tmp/cache/meta",
    :entitystore => "file:tmp/cache/body"
  )
  
  def self.mime(ext, type)
    ext = ".#{ext}" unless ext.to_s[0] == ?.
    Rack::Mime::MIME_TYPES[ext.to_s] = type
  end
    
  mime(:json, "application/json")
  mime(:png, "image/png")
  mime(:svg, "image/svg+xml")

  error do
    error = request.env['sinatra.error']
    response["Content-Type"] = "text/plain;charset=UTF-8"
    
    if ENV["NOMS_DEBUG"]
      error.to_s + ": " + error.backtrace.join("\n")
    else
      error.to_s
    end
  end
end
