# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2010  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class NOMS::WebApp
  get("/") do
    cached()
    @groups = sort_by_group(NOMS.servers.values)    
    erb :index
  end
  
  NOMS.servers.each_value do |server_config|
    server = server_config.config.short_name
    
    get("/#{server}/?") do
      setup_server(server)
      servers = NOMS.config.servers_grouped.map { |short_name| NOMS.servers[short_name] }
      @groups = sort_by_group(servers).map { |(group, servers)|
        [group, servers.map { |s| server_link(s.config, false) }]
      }
      cached()
      erb :server_index
    end

    if server_config.config.show_expanded_datasets
      get("/#{server}/match_history") do
        setup_server(server)
        erb :match_history
      end

      get("/#{server}/match_history/player") do
        setup_server(server)
        wins = NOMS::DB::Player.formatted_match_wins(params[:name])
        content_type(:json)
        wins.to_json
      end
      
      get("/#{server}/achievements") do
        setup_server(server)

        @should_display_player_data = false

        @number_matches_won_all_time = NOMS::DB::MatchWin.distinct_matches_count
        @number_players_all_time = NOMS::DB::Player.count

        @achievements = Hash.new { |h,k| h[k] = {} }

        @achievements["perfect_match"]["number_matches"] = NOMS::DB::PerfectMatchAchievement.count
        @achievements["perfect_match"]["number_distinct_players"] = NOMS::DB::PerfectMatchAchievement.distinct_players_count
        @achievements["perfect_match"]["last_n"] = NOMS::DB::PerfectMatchAchievement.order_by(:global_match_number).last(5)

        @achievements["minimum_match"]["number_matches"] = NOMS::DB::MinimumMatchAchievement.distinct_matches_count
        @achievements["minimum_match"]["number_distinct_players"] = NOMS::DB::MinimumMatchAchievement.distinct_players_count
        @achievements["minimum_match"]["last_n"] = NOMS::DB::MinimumMatchAchievement.order_by(:global_match_number).last(5)

        @achievements["round_loser"]["number_matches"] = NOMS::DB::MatchWinnerRoundLoserAchievement.count
        @achievements["round_loser"]["number_distinct_players"] = NOMS::DB::MatchWinnerRoundLoserAchievement.distinct_players_count
        @achievements["round_loser"]["last_n"] = NOMS::DB::MatchWinnerRoundLoserAchievement.order_by(:global_match_number).last(5)

        @achievements["round_winner"]["number_matches"] = NOMS::DB::MatchWinnerRoundWinnerAchievement.count
        @achievements["round_winner"]["number_distinct_players"] = NOMS::DB::MatchWinnerRoundWinnerAchievement.distinct_players_count
        @achievements["round_winner"]["last_n"] = NOMS::DB::MatchWinnerRoundWinnerAchievement.order_by(:global_match_number).last(5)

        @achievements["domination"]["number_matches"] = NOMS::DB::DominationAchievement.count
        @achievements["domination"]["number_distinct_players"] = NOMS::DB::DominationAchievement.distinct_players_count
        @achievements["domination"]["last_n"] = NOMS::DB::DominationAchievement.order_by(:global_match_number).last(5)

        calculate_common_achievement_stats(@achievements, @number_matches_won_all_time, @number_players_all_time)

        cached()
        erb :achievements
      end
      
      if NOMS::WebApp::StatSignature.supported
        get("/#{server}/statsig") do
          setup_server(server)
          @signature = NOMS::WebApp::StatSignature.new(nil)
          @name = params[:name] || ""
          erb :statsig
        end

        get("/#{server}/statsig/generate.png") do
          setup_server(server)

          cached()
          player = NOMS::DB::Player.find_by_name(params[:name])
          signature = NOMS::WebApp::StatSignature.new(player)

          signature.check_option(params, :font) { |font| NOMS::WebApp::StatSignature.font(font) }
          signature.check_option(params, :font_pointsize) { |size| Float(size) }
          signature.check_option(params, :text_antialias) { |antialias| antialias == "true" }
          signature.check_option(params, :background_image) { |uri| uri }
          signature.check_option(params, :fill_color) { |color| color }
          signature.check_option_coord(params, :name_pos) { |c| Integer(c) }
          signature.check_option_coord(params, :server_pos) { |c| Integer(c) }
          signature.check_option_coord(params, :basic_stats_pos) { |c| Integer(c) }

          content_type(:png)
          signature.to_png
        end
      end
    end
    
    get("/#{server}/player/?") do
      setup_server(server)

      @players = []
      @should_display_player_data = false
      @achievements = Hash.new { |h,k| h[k] = {} }
      @title = "Player Data"

      if params[:name] && params[:name].any?
        p = NOMS::DB::Player[:name => params[:name]]
        @players = [p] if p
      elsif params[:q] && params[:q].any?
        @query = params[:q]
        @players = NOMS::DB::Player.search_players(@query).all
      end

      if [1, 2].include?(@players.size)
        @should_display_player_data = true
        player = @players[0]
        @title += " for #{player.display_name}"
        @stats_table = StatsTable.new

        @achievements["perfect_match"]["last_n"] = player.perfect_match_achievements
        @achievements["perfect_match"]["number_matches"] = @achievements["perfect_match"]["last_n"].length

        @achievements["minimum_match"]["last_n"] = player.minimum_match_achievements
        @achievements["minimum_match"]["number_matches"] = @achievements["minimum_match"]["last_n"].length

        @achievements["round_loser"]["last_n"] = player.match_winner_round_loser_achievements
        @achievements["round_loser"]["number_matches"] = @achievements["round_loser"]["last_n"].length

        @achievements["round_winner"]["last_n"] = player.match_winner_round_winner_achievements
        @achievements["round_winner"]["number_matches"] = @achievements["round_winner"]["last_n"].length

        @achievements["domination"]["last_n"] = player.domination_achievements
        @achievements["domination"]["number_matches"] = @achievements["domination"]["last_n"].length
      end

      @title = h(@title)
      erb :player
    end
    
    get("/#{server}/stats") do
      setup_server(server)
    
      # Disallow a user from requesting a new page that is not cached by using spurious parameters.
      validate_params_keys(%w[tier], "/#{server}/stats")
    
      @tier_range = (1..3)
      if params[:tier]
        @tier = params[:tier].to_i
        if @tier <= 1 || @tier > @tier_range.end
          redirect("/#{server}/stats", 303)
        end
      else
        @tier = 1
      end
    
      stats_dataset = NOMS::DB::Player.stats_data(@tier)
      @record_range = stats_dataset.current_page_record_range
    
      cached()
      @stats_table = StatsTable.new
      @stats = stats_dataset.all
      erb :stats
    end
    
    if server_config.config.armaskill.generate
      get("/#{server}/stats/best") do
        setup_server(server)
        validate_params_keys([], "/#{server}/stats/best")
        cached()
        @stats_table = RatingsStatsTable.new
        @stats = NOMS::DB::Player.ratings_data.all
        erb :stats_best
      end
    end
    
    get("/#{server}/compare") do
      setup_server(server)
      @base_player = NOMS::DB::Player[:name => "dlh@generalconsumption.org"]
      @stats_table = CompareStatsTable.new(@base_player)
      @stats = [@base_player, NOMS::DB::Player[:name => "player_1"]]
      erb :compare
    end
    
    get("/#{server}/world-map-players.json") do
      setup_server(server)
      map = NOMS::Choropleth::Map.new
      cached()
      content_type(:json)
      map.to_json
    end
    
  end
end