# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2010  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "noms/util"

class NOMS::WebApp::StatSignature
  class << self
    attr_accessor :supported
  end
end

begin
  require "RMagick"
  NOMS::WebApp::StatSignature.supported = true
rescue LoadError => e
  NOMS::WebApp::StatSignature.supported = false
  warn("Warning #{e}: StatSignature support not available")
end

require "net/http"
require "uri"

class NOMS::WebApp::StatSignature
  class Error < StandardError; end
  MAX_FETCH_SIZE = 128_000
  
  Coord = Struct.new(:x, :y)
  
  FONTS = {
    "silkscreen" => "'#{NOMS::Util.public_path("fonts/slkscre.ttf")}'",
    "Armagetronad" => "'#{NOMS::Util.public_path("fonts/Armagetronad.ttf")}'",
    "DejaVuSansMono" => "'#{NOMS::Util.public_path("fonts/DejaVuSansMono.ttf")}'",
    "DejaVuSerif" => "'#{NOMS::Util.public_path("fonts/DejaVuSerif.ttf")}'"
  }
  
  def self.font(name)
    FONTS[name] || FONTS["silkscreen"]
  end
  
  def initialize(player)
    @player = player
    @name_pos = Coord.new(20, 16)
    @server_pos = Coord.new(16, 83)
    @basic_stats_pos = Coord.new(100, 35)
    @background_image = "/stylesheets/img/statsig.png"
    @font = self.class.font("DejaVuSansMono")
    @font_pointsize = 10
    @text_antialias = true
    @fill_color = "black"
  end
  
  def check_option(params, option)
    if params[option] && params[option].any?
      self.method(:"#{option}=").call(yield(params[option]))
    end
  end
  
  def check_option_coord(params, option)
    x_option = "#{option}_x".to_sym
    y_option = "#{option}_y".to_sym
    
    if params[x_option] && params[x_option].any?
      self.method(option).call.x = yield(params[x_option])
    end
    if params[y_option] && params[y_option].any?
      self.method(option).call.y = yield(params[y_option])
    end
  end
  
  attr_accessor :background_image
  attr_accessor :font
  attr_accessor :font_pointsize
  attr_accessor :text_antialias
  attr_accessor :fill_color
  attr_reader :name_pos
  attr_reader :server_pos
  attr_reader :basic_stats_pos
  
  def to_png
    background_image = Magick::Image.from_blob(fetch_background())[0]
    image = Magick::Image.new(background_image.columns, background_image.rows)
    image.opacity = Magick::MaxRGB
    draw = Magick::Draw.new
    # draw.stroke("transparent")
    draw.fill(@fill_color)
    draw.pointsize(@font_pointsize)
    draw.text_antialias(@text_antialias)
    draw.font(@font)
    draw.text(@name_pos.x, @name_pos.y, @player.display_name)
    draw.text(@server_pos.x, @server_pos.y, NOMS.config.name)
    build_table(image, draw, @basic_stats_pos, [
      [" ", "Matches", "Rounds"],
      [" ", ordinalize(@player.won_matches_rank), ordinalize(@player.won_rounds_rank)],
      ["Won", format_number_human(@player.won_matches), format_number_human(@player.won_rounds)],
      ["% Won", to_percent(@player.matches_percent_won), to_percent(@player.rounds_percent_won)]
    ])
    draw.draw(image)
    
    background_image.composite!(image, Magick::CenterGravity, Magick::OverCompositeOp)
    background_image.format = "PNG"
    background_image.to_blob
  end
  
  protected
  
  def fetch_background
    uri = URI.parse(@background_image)
    if uri.scheme != "http"
      NOMS::Util.check_path(@background_image)
      return File.read(NOMS::Util.public_path(@background_image))
    end
    
    Net::HTTP.start(uri.host, uri.port) do |http|
      head = http.head(uri.path)
      if Integer(head["Content-Length"]) > MAX_FETCH_SIZE
        raise Error.new("Background image too large. #{head["Content-Length"]} bytes > #{MAX_FETCH_SIZE} bytes")
      end
      res = http.get(uri.path)
      if res.code != "200"
        raise Error.new("Couldn't fetch background image at URI #{@background_image} (#{res.code})")
      end
      res.body
    end
  end
  
  def build_table(image, draw, pos, columns)
    column_widths = columns.map { |rows| 
      max_str = rows.max { |a, b| draw.get_type_metrics(image, a).width <=> draw.get_type_metrics(image, b).width }
      draw.get_type_metrics(image, max_str).width + 7
    }
    
    draw.text_align(Magick::LeftAlign)
    dy = draw.get_type_metrics("foo").height
    y = pos.y
    unzip(columns, columns[0].length).each do |row|
      draw.line(pos.x, y + 3, column_widths.inject(pos.x) { |memo, w| memo + w }, y + 3)
      x = pos.x
      row.each_with_index do |s, index|
        draw.text(x, y, s)
        x += column_widths[index] + 5
      end
      y += dy
    end
    
    x = pos.x
    column_widths.each_with_index do |width, index|
      draw.line(x, pos.y - dy + 7, x, pos.y + (dy * (columns[0].length - 1)) + 3) if index > 0
      x += width + 3
      x = x.to_i
    end
    
  end
  
  def unzip(xs, n)
    (0...n).map { |i| xs.map { |tuple| tuple[i] } }
  end
end
