# NOMS -- A web-app for Armagetron Advanced statistics
# Copyright (C) 2009  Daniel Harple
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation in version 3 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "enumerator"

class NOMS::WebApp::StatsTable
  include NOMS::WebApp::Helpers
  
  Group = Struct.new(:name, :options)
  class Group
    def initialize(name)
      @name = name
      @options = []
      @stats_table = "stats_table"
      yield self if block_given?
    end
    attr_reader :name
    attr_reader :options
    attr_accessor :stats_table
    
    def add_option(name, description, player_method_sym, smaller_is_better=false, &formatter)
      @options << Option.new(self, name, description, player_method_sym, smaller_is_better, &formatter)
    end
  end
  
  class Option
    def initialize(group, name, description, player_method_sym, smaller_is_better=false, &formatter)
      @group = group
      @name = name
      @player_method_sym = player_method_sym
      @description = description
      @smaller_is_better = smaller_is_better
      @formatter = formatter
    end
    attr_accessor :name, :description
    attr_writer :enabled

    def call(player)
      return format(player.method(@player_method_sym).call)
    end
    
    def smaller_is_better?
      @smaller_is_better
    end
    
    def enabled?
      @enabled
    end
    
    def diff(player1, player2)
      value = player1.method(@player_method_sym).call - player2.method(@player_method_sym).call
      prefix = value > 0 ? "+" : ""
      return prefix + format(value)
    end
    
    private
    
    def format(value)
      return value.to_s if !@formatter
      return @formatter.call(value)
    end
  end

  def initialize(should_check_config=true)
    @default_sorting_order = ["Matches", "Won"]
    @should_display_name_and_rank = true
    
    @groups = []
    @groups << Group.new("Matches") { |g|
      g.add_option("Won", "The amount of won matches.", :won_matches_for_display)
      g.add_option("% Won", "The percentage of matches you win when you play.", :matches_percent_won) { |n| to_percent(n) }
      g.add_option("Played", "The total number of matches you have played.", :matches_played)
      # g.add_option("% Won Global", "Using the global average match length, the percentage you win when you play.", false) { |p| round(p.matches_percent_won_global) }
      # g.add_option("Win Streak", "The greatest number of matches you have won in a strictly sequential row (no breaks in-between matches).") { |p| p.match_win_streak_strict }
    }
    @groups << Group.new("Detailed Matches") { |g|
      g.add_option("Won by Avg.", "The average number of points you win a match by.", :won_by_avg) { |n| round(n) }
      g.add_option("Won by Max", "The greatest number of points you won a match by.", :won_by_max)
      g.add_option("To Win Avg.", "The average number of rounds it takes you to win a match.", :matches_average_number_rounds, true) { |n| round(n) }
      # g.add_option("Played Avg.", "The average number of rounds you play in a match.") { |p| round(p.average_match_length) }
      # g.add_option("To Win Min", "The least number of rounds you played to win a match.", true) { |p| p.matches_min_number_rounds }
      # g.add_option("To Win Max", "The greatest number of rounds you played to win a match.", true) { |p| p.matches_max_number_rounds }
    }
    @groups << Group.new("Rounds") { |g|
      g.add_option("Won", "The amount of won rounds.", :won_rounds_for_display)
      g.add_option("% Won", "The percentage of rounds you win when you play.", :rounds_percent_won) { |n| to_percent(n) }
      g.add_option("Played", "The total amount of rounds you have played.", :rounds_played)
    }
    @groups << Group.new("Points") { |g|
      g.add_option("Total", "The total amount of points you have scored.", :round_points_total)
      g.add_option("Avg.", "The average points you score per round.", :points_average) { |n| round(n) }
      g.add_option("% Zero", "The percentage of rounds you acquire no points.", :points_percent_zero, true) { |n| to_percent(n) }
      g.add_option("Round Max", "The most points you have scored in one round.", :round_points_highscore)
      g.add_option("Match Max", "The most points you have scored in a match.", :match_points_highscore)
    }
    @groups << Group.new("Team Points") { |g|
      g.add_option("Total", "The total amount of points your team has scored.", :team_round_points_total)
      g.add_option("Enemies", "The total summation of points your opponents have scored.", :opponents_round_points_total, true)
      g.add_option("Avg.", "The average points your team scores per round.", :team_points_average) { |n| round(n) }
      g.add_option("Enemies Avg.", "The average points your opponents score per round.", :opponents_points_average, true) { |n| round(n) }
      g.add_option("Diff", "The difference between your team points average and opponents points average.", :team_points_diff) { |n| round(n) }
      g.add_option("Round Max", "The most points your team has scored in a round.", :team_round_points_highscore)
      g.add_option("Match Max", "The most points your team has scored in a match.", :team_match_points_highscore)
    }
    @groups << Group.new("ArmaSkill") { |g|
      g.add_option("Rating", "A conservative estimate of your skill. This rating system is an implementation of Microsoft's <a href=\"http://research.microsoft.com/en-us/projects/trueskill\">TrueSkill</a>.", :rating) { |n| round(n) }
      g.add_option("Mean", "The average skill you have.", :mu) { |n| round(n) }
      g.add_option("Standard Deviation", "The degree of uncertainty in your skill.", :sigma, true) { |n| round(n) }
    }
    @groups << Group.new("Core Dumps") { |g|
      g.add_option("Hunter Total", "The total amount of times you have core dumped another player", :hunter_total)
      g.add_option("Dumped Total", "The total amount of times you have been core dumped", :prey_total, true)
      g.add_option("Hunter Avg.", "The average number of players you core dump per round", :core_dumps_hunter_average) { |n| round(n) }
      g.add_option("% Dumped", "The percentage you are core dumped per number of rounds played.", :core_dumps_percent_dumped, true) { |n| to_percent(n) }
      g.add_option("Kills Per Death", "The ratio of the total amount of core dumps and the total amount of deaths (suicides, teamkills, and total dumped).", :kill_death_ratio) { |n| round(n) }
    }
    @groups << Group.new("Team Kills") { |g|
      g.add_option("Team Killed", "The total amount of times you have been team killed.", :teamkilled_total, true)
      g.add_option("Team Killer", "The total amount of times you have team killed a team mate.", :teamkiller_total, true)
      g.add_option("% Team Killed", "The percentage of rounds you are team killed.", :teamkilled_average, true) { |n| to_percent(n) }
      g.add_option("Team Killer Avg.", "The average number of times you team kill a team mate per round.", :teamkiller_average, true) { |n| round(n) }
    }
    @groups << Group.new("Suicides") { |g|
      g.add_option("Total", "The total amount of suicides.", :suicides_total, true)
      g.add_option("% Suicided", "The percentage you suicide per number of rounds played.", :suicides_percent_suicided, true) { |n| to_percent(n) }
    }
    @groups << Group.new("Zones") { |g|
      g.add_option("Won", "The total amount of zones that you have helped conquered.", :basezone_conqueror)
      g.add_option("Team Lost", "The total amount of your team's zones that have been conquered.", :basezone_conquered, true)
      g.add_option("Lost", "The total amount of your team's zones that have been conquered while you are alive and in the zone.", :basezone_conqueror_defender)
      g.add_option("Won Avg.", "The average number of zones you have helped conquered per the number of rounds you have played.", :basezone_conqueror_avg) { |n| round(n) }
      g.add_option("Team Lost Avg.", "The average number of your team's zones that have been conquered per the number of rounds you have played.", :basezone_conquered_avg, true) { |n| round(n) }
      g.add_option("Lost Avg.", "The average number of your team's zones that have been conquered while you are alive and in the zone per number of rounds you have played.", :basezone_conqueror_defender_avg) { |n| round(n) }
    }
    @groups << Group.new("Target Zones") { |g|
      g.add_option("Attempts", "The total amount of attempts you have made.", :cycles_created)
      g.add_option("Completions", "The total amount of target zones you have reached.", :target_zones_entered)
      g.add_option("% Completed", "The percentage you reach a target zone.", :target_zone_completion_percent) { |n| to_percent(n) }
      g.add_option("Avg.", "The average number of target zones you reach per round.", :target_zone_avg) { |n| round(n) }
    }
    @groups << Group.new("Sacrifices") { |g|
      g.add_option("Holes Used", "The total number of holes you have used.", :holes_used_total)
      g.add_option("Holes Made", "The total number of holes you have made which have been used.", :holes_made_total, true)
      g.add_option("Holed", "The total number of times your wall has been holed.", :holed_total, true)
      g.add_option("Holes Used Avg.", "The average number of holes you use per rounds played.", :holes_used_average) { |n| round(n) }
      g.add_option("Holes Made Avg.", "The average number of holes you make per rounds played", :holes_made_average, true) { |n| round(n) }
      g.add_option("Holed Avg.", "The average number of times your wall is holed per rounds played.", :holed_total_average, true) { |n| round(n) }
    }
    # @groups << Group.new("Achievements") { |g|
    #   g.add_option("Minimum", "The amount of Minimum Match Achievements you have") { |p| p.minimum_match_achievements_dataset.count }
    #   g.add_option("Dominating", "The amount of Dominating Match Achievements you have") { |p| p.domination_achievements_dataset.count }
    #   g.add_option("Perfect", "The amount of Perfect Match Achievements you have.") { |p| p.perfect_match_achievements_dataset.count }
    #   g.add_option("Winner", "The amount of Match Winner, Round Winner Achievements you have") { |p| p.match_winner_round_winner_achievements_dataset.count }
    #   g.add_option("Loser", "The amount of Match Winner, Round Loser Achievements you have") { |p| p.match_winner_round_loser_achievements_dataset.count }
    # }

    check_config(false) if should_check_config
  end

  attr_accessor :should_display_name_and_rank
  attr_reader :groups

  def transform_stats_table_header_name(header)
    header.downcase.gsub(" ", "_")
  end

  def transform_stats_table_option_name(header, name)
    header = transform_stats_table_header_name(header)
    name = name.downcase.gsub("%", "percent").gsub(" ", "_").gsub("/", "slash")
    name = name.gsub(".", "")

    return header + "-" + name
  end

  def column_headers1_attributes(group)
    number_options_enabled = group.options.inject(0) { |n, option| option.enabled? ? n + 1 : n }
    attrs = {}
    attrs["id"] = transform_stats_table_option_name(group.name, "header")
    attrs["colspan"] = number_options_enabled
    attrs["class"] = transform_stats_table_header_name(group.name)
    attrs["class"] += " disabled hide" if number_options_enabled.zero?
    return attrs
  end

  def column_class_attributes(group, option)
    class_attrs = [transform_stats_table_header_name(group.name), transform_stats_table_option_name(group.name, option.name)]
    class_attrs += %w[disabled hide] if !option.enabled?
    return class_attrs.join(" ")
  end
  
  def smaller_is_better_option_names
    self.groups.inject([]) { |memo, group|
      group.options.each do |opt|
        memo << transform_stats_table_option_name(group.name, opt.name) if opt.smaller_is_better?
      end
      memo
    }
  end
  
  def default_sorting_order_column
    opt_index = 0
    @groups.each_with_index do |group, index|
      if group.name == @default_sorting_order[0]
        opt_index += group.options.enum_for(:each_with_index).find { |opt, index| opt.name == @default_sorting_order[1] }[1]
        break
      else
        opt_index += group.options.length
      end
    end
    opt_index += 2 if @should_display_name_and_rank
    opt_index
  end
  
  protected
  
  def check_config(default_option_enabled)
    @groups.each do |group|
      if NOMS.config[group.stats_table][group.name].enabled != "default"
        group.options.each do |opt|
          opt.enabled = NOMS.config[group.stats_table][group.name].enabled
        end
      else
        group.options.each do |opt|
          opt.enabled = if NOMS.config[group.stats_table][group.name][opt.name] != "default"
            NOMS.config[group.stats_table][group.name][opt.name]
          else
            default_option_enabled
          end
        end
      end
    end
    
    @groups.delete_if do |group|
      !NOMS.config.stats_table[group.name].available
    end
  end
  
  
end

class NOMS::WebApp::RatingsStatsTable < NOMS::WebApp::StatsTable
  def initialize(should_check_config=true)
    super(false)
    
    @default_sorting_order = ["ArmaSkill", "Rating"]
    
    @groups.each do |group|
      group.stats_table = "ratings_stats_table"
    end
    
    check_config(false) if should_check_config
  end
end

class NOMS::WebApp::CompareStatsTable < NOMS::WebApp::StatsTable
  module CompareOption
    attr_accessor :base_player
    def call(player)
      return old_call(player) if player == self.base_player
      return diff(player, self.base_player)
    end
  end
  
  def initialize(base_player, should_check_config=true)
    super(false)
    
    @base_player = base_player
    
    @groups.each do |group|
      group.stats_table = "compare_stats_table"
      
      group.options.each do |opt|
        class << opt
          alias_method(:old_call, :call)
          include CompareOption
        end
        opt.base_player = @base_player
      end
    end
    
    check_config(false) if should_check_config
  end
  attr_reader :base_player
end
