require "test/unit"
require "noms/config"

class TestConfig < Test::Unit::TestCase
  def setup
    @config = NOMS::Config.new
  end
  
  def test_simple_top_level
    @config.add_valid_option("foo", "")
    @config.from_hash("foo" => 1)
    assert_equal(1, @config.foo)
  end
  
  def test_simple_default
    @config.add_valid_option("foo", "", 2)
    assert_equal(2, @config.foo)
  end
  
  def test_simple_default_set
    @config.add_valid_option("foo", "", 2)
    @config.from_hash("foo" => 1)
    assert_equal(1, @config.foo)
  end
  
  def test_nested
    @config.new_nested_option("bar", "") do |c|
      c.add_valid_option("baz", "")
    end
    @config.from_hash({"bar" => {"baz" => 2}})
    assert_equal(2, @config.bar.baz)
  end
  
  def test_nested_default
    nested = @config.new_nested_option("foo", "")
    nested.add_valid_option("bar", "", 1)
    assert_equal(1, @config.foo.bar)
  end
  
  def test_even_more_nested
    nested1 = @config.new_nested_option("foo", "")
    nested2 = nested1.new_nested_option("bar", "")
    nested2.add_valid_option("baz", "", 1)
    nested1.add_valid_option("bing", "", 2)
    assert_equal(1, @config.foo.bar.baz)
    assert_equal(2, @config.foo.bing)
  end
  
  def test_validator
    @config.add_valid_option("foo", "") { |n| n > 0 }
    @config.from_hash("foo" => 1)
    assert_equal(1, @config.foo)
  end
  
  def test_validator_error
    @config.add_valid_option("foo", "") { |n| n > 0 }
    assert_raise(NOMS::Config::Error) { @config.from_hash("foo" => -1) }
  end
  
  def test_import_error
    assert_raise(NOMS::Config::Error) { @config.from_hash("not valid" => 0) }
  end
  
  def test_stats_table_nested
    @config.add_valid_option("name", "") { |value| value && value.any? }
    
    @config.new_nested_option("stats_table", "") do |c|
      c.new_nested_option("ArmaSkill", "") do |group_config|
        group_config.add_valid_option("enable_group", "", true)
        group_config.add_valid_option("Points", "", true)
      end
      c.new_nested_option("Points", "") do |group_config|
        group_config.add_valid_option("enable_group", "", true)
      end
    end
    @config.from_hash("stats_table" => {"ArmaSkill" => {"enable_group" => false}}, "name" => "foo")
    assert_equal(false, @config.stats_table.ArmaSkill.enable_group)
    assert_equal(true, @config.stats_table.Points.enable_group)
    assert_equal("foo", @config.name)
  end
  
  def test_index_operator
    @config.add_valid_option("foo", "", 3)
    assert_equal(3, @config["foo"])
  end
  
  def test_multiple_imports
    @config.add_valid_option("foo", "")
    @config.add_valid_option("bar", "")
    
    @config.from_hash("foo" => 0, "bar" => 1)
    assert_equal(0, @config.foo)
    assert_equal(1, @config.bar)
    
    @config.from_hash("foo" => 2)
    assert_equal(1, @config.bar)
    assert_equal(2, @config.foo)
  end
  
  def test_import_base_config
    @config.add_valid_option("foo", "")
    @config.add_valid_option("bar", "")
    @config.add_valid_option("baz", "")
    @config.add_valid_option("bing", "", "default")
    
    # Import the base config
    # Normally this would throw an error because config.baz needs to be set
    @config.from_hash({"foo" => "base", "bar" => "base"}, false)
    
    # Import the normal config
    @config.from_hash({"foo" => "overloaded", "baz" => "initially set"})
    
    assert_equal("overloaded", @config.foo)
    assert_equal("base", @config.bar)
    assert_equal("initially set", @config.baz)
    assert_equal("default", @config.bing)
  end
end
