require "test/unit"
require "noms"

NOMS::Setup.add_config("config/noms-examples/mud-puddle.config")
NOMS::Setup.generate_models(false, false)

class TestLadderlog < Test::Unit::TestCase
  include NOMS::Ladderlog::Action
  
  def test_xdude_match_win
    match = load_match("xdude_match_win.txt")
    assert(match.completed?)
    assert_equal("xdude", match.winner.team)
  end
  
  def test_authenticated_match_win
    match = load_match("authenticated_match_win.txt")
    assert(match.completed?)
    assert_equal("free_kill", match.winner.team)
    assert(match.teams[match.winner.team].players.include?("dlh@generalconsumption.org"))
  end
  
  def test_single_player_ai
    match = load_match("single_player_ai.txt")
    assert(!match.multiplayer?)
  end
  
  def test_single_player_ai2
    match = load_match("single_player_ai2.txt")
    assert(!match.multiplayer?)
  end

  def test_team_game_won_by
    match = load_match("team_game_match_win.txt")
    assert(match.completed?)
    assert_equal("the_usual_suspects", match.winner.team)
    assert_equal(46, match.__send__(:match_won_by_amount))
  end
  
  def test_basic
    action = Base.new("ROUND_WINNER me")
    assert_equal("ROUND_WINNER", action.name)
    assert_equal("me", action.line)
    assert_equal(["me"], action.parse_regular_tokens)
  end
  
  def test_special_token
    action = Base.new("SPECIAL regular1 regular2 foo bar")
    assert_equal("foo bar", action.parse_special_token(2))
  end  
  
  def test_special_token2
    action = Base.new("SPECIAL regular1 regular2  foo bar ")
    assert_equal(" foo bar ", action.parse_special_token(2))
  end
  
  def test_regular_token_empty
    action = Base.new("FOO bar  baz")
    assert_equal(["bar", "", "baz"], action.parse_regular_tokens)
  end
  
  def test_round_score
    action = with_action(RoundScore, "ROUND_SCORE 3 free_kill free_kill")
    assert_equal(3, action.score)
    assert_equal("free_kill", action.player)
    assert_equal("free_kill", action.team)
  end
  
  def test_round_score_spectator
    action = with_action(RoundScore, "ROUND_SCORE 0 free_kill")
    assert_equal("", action.team)
  end
  
  def test_targetzone_enter
    action = with_action(TargetzonePlayerEnter, "TARGETZONE_PLAYER_ENTER 0  12.5 56.25 free_kill 12.5359 51.3915 0 1 6.44212")
    assert_equal("free_kill", action.player)
  end
  
  def test_player_entered_ip_stripped_screen_name
    action = with_action(PlayerEntered, "PLAYER_ENTERED free_kill free kill", false)
    assert_equal("free kill", action.screen_name)
  end
  
  def test_player_entered_no_screen_name
    action_with_ip_addresses = with_action(PlayerEntered, "PLAYER_ENTERED ca$h 192.168.1.1", true)
    action_without_ip_addresses = with_action(PlayerEntered, "PLAYER_ENTERED ca$h", false)
    assert_equal("", action_with_ip_addresses.screen_name)
    assert_equal("", action_without_ip_addresses.screen_name)
  end
  
  def test_player_entered_screen_name
    action_with_ip_addresses = with_action(PlayerEntered, "PLAYER_ENTERED free_kill 0.0.0.0 free kill", true)
    action_without_ip_addresses = with_action(PlayerEntered, "PLAYER_ENTERED free_kill free kill", false)
    assert_equal("free kill", action_with_ip_addresses.screen_name)
    assert_equal("free kill", action_without_ip_addresses.screen_name)
  end
  
  def test_player_renamed_ip_stripped
    action = with_action(PlayerRenamed, "PLAYER_RENAMED free_kill dlh@generalconsumption.org free kill", false, false)
    assert_equal("free kill", action.screen_name)
    assert_equal("free_kill", action.old_name)
    assert_equal("dlh@generalconsumption.org", action.new_name)
    assert_equal(true, action.authenticated_rename?)
  end
  
  def test_player_renamed_sty_ct
    action = with_action(PlayerRenamed, "PLAYER_RENAMED free_kill dlh@generalconsumption.org 0.0.0.0 1 free kill", true, true)
    assert_equal("free kill", action.screen_name)
    assert_equal("free_kill", action.old_name)
    assert_equal("dlh@generalconsumption.org", action.new_name)
    assert_equal(true, action.authenticated_rename?)
  end
  
  def test_player_renamed_sty_ct2
    action = with_action(PlayerRenamed, "PLAYER_RENAMED ct_kyle kyle@ct/jedi 0.0.0.0 1 ct¤kyle", true, true)
    assert_equal("ct¤kyle", action.screen_name)
  end
  
  def test_player_renamed_no_screen_name
    action_with_ip_addresses = with_action(PlayerRenamed, "PLAYER_RENAMED td|l td|longball 24.245.62.241", true, false)
    action_without_ip_addresses = with_action(PlayerRenamed, "PLAYER_RENAMED td|l td|longball", false, false)
    
    assert_equal("", action_with_ip_addresses.screen_name)
    assert_equal("", action_without_ip_addresses.screen_name)
  end
  
  def test_player_renamed_screen_name
    action_with_ip_addresses = with_action(PlayerRenamed, "PLAYER_RENAMED free_kill dlh@generalconsumption.org 0.0.0.0 free kill", true, false)
    action_without_ip_addresses = with_action(PlayerRenamed, "PLAYER_RENAMED free_kill dlh@generalconsumption.org free kill", false, false)
    assert_equal("free kill", action_with_ip_addresses.screen_name)
    assert_equal("free kill", action_without_ip_addresses.screen_name)
  end
  
  def test_match_winner_no_players
    action = with_action(MatchWinner, "MATCH_WINNER team_gold")
    assert_equal("team_gold", action.team)
    assert_equal([], action.players)
  end
  
  def test_match_winner_players
    action = with_action(MatchWinner, "MATCH_WINNER team_gold p1 p2")
    assert_equal("team_gold", action.team)
    assert_equal(["p1", "p2"], action.players)
  end
  
  protected
  
  def load_matches(path)
    File.open("test/fixtures/#{path}") do |f|
      NOMS::Ladderlog::Parser.new(f, 0).all_matches
    end
  end
  
  def load_match(path)
    matches = load_matches(path)
    assert(matches.length > 0)
    matches[0]
  end
  
  def with_action(klass, *args)
    base = Base.new(args[0])
    klass.new(base, *args[1..-1])
  end
end
